# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Json utils."""

from typing import Mapping
import json
from typing import Any

import numpy as np
from vizier import pyvizier as vz


class NumpyEncoder(json.JSONEncoder):
  """Example: json.dumps(np.array(...), cls=NumpyEncoder)."""

  def default(self, o: Any) -> Any:
    if not isinstance(o, dict) and isinstance(o, Mapping):
      # Handles FrozenDict and any dict-like structures.
      return dict(o)
    elif set(dir(o)).issuperset(set(['tolist', 'dtype', 'shape'])):
      # Any numpy.array-like objects.
      # Shape must be dumped and restored, in case that the array has shape
      # that includes 0-sized dimensions.
      return {
          'dtype': np.dtype(o.dtype).name,
          'value': o.tolist(),
          'shape': o.shape
      }
    else:
      return o


def numpy_hook(obj: Any) -> Any:
  """Example: json.loads(..., object_hook=numpy_hook)."""
  if 'dtype' not in obj:
    return obj
  if 'shape' not in obj:
    return obj
  return np.array(obj['value'], dtype=obj['dtype']).reshape(obj['shape'])


class NumpyDecoder(json.JSONDecoder):
  """Convert list to numpy if we see a list during json decoding.

  Example: json.loads(json_string, cls=NumpyDecoder).
  """

  def __init__(self, *args, **kargs):
    super(NumpyDecoder, self).__init__(object_hook=numpy_hook, *args, **kargs)


class MetadataEncoder(json.JSONEncoder):

  def default(self, o: Any) -> Any:
    if isinstance(o, vz.Metadata):
      d = dict(o)
      d |= {ns.encode(): o.ns(ns) for ns in o.subnamespaces()}
      return d
    elif not isinstance(o, dict) and isinstance(o, Mapping):
      # Handles FrozenDict and any dict-like structures.
      return dict(o)
    return o



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Numpy experimenter for wrapping deterministic functions on ndarrays."""

import copy
import logging
import math
from typing import Callable, Sequence

import numpy as np
from vizier import pyvizier
from vizier._src.benchmarks.experimenters import experimenter
from vizier.pyvizier import converters


def _get_name(f):
  """Gets the name of underlying objects."""
  if hasattr(f, '__name__'):
    return f.__name__
  # Next clause handles functools.partial objects.
  if hasattr(f, 'func') and hasattr(f.func, '__name__'):
    return f.func.__name__
  return repr(f)


class NumpyExperimenter(experimenter.Experimenter):
  """NumpyExperimenters take a deterministic function on ndarrays."""

  def __init__(
      self,
      impl: Callable[[np.ndarray], float],
      problem_statement: pyvizier.ProblemStatement,
  ):
    """NumpyExperimenter with analytic function impl for one metric.

    NumpyExperimenter only supports single objectives, and flat numeric search
    spaces.

    Args:
      impl: Function that scalarizes np.ndarray of shape (dimension,).
      problem_statement: Problem statement.

    Raises:
      ValueError: Non-positive dimension or invalid problem statement.
    """
    dimension = len(problem_statement.search_space.parameters)
    self._impl_name = _get_name(impl)
    logging.info(
        'Initializing NumpyExperimenter with impl=%s, dimension=%s',
        self._impl_name,
        dimension,
    )
    if dimension <= 0:
      raise ValueError(f'Invalid dimension: {dimension}')
    self._dimension = dimension
    self.impl = impl

    if not problem_statement.metric_information.is_single_objective:
      raise ValueError(
          f'Statement should be single objective {problem_statement}'
      )
    if problem_statement.search_space.is_conditional:
      raise ValueError(f'Statement should be flat {problem_statement}')
    for parameter in problem_statement.search_space.parameters:
      if not parameter.type.is_numeric():
        raise ValueError(f'Non-numeric parameters {parameter}')

    objective_metrics = problem_statement.metric_information.of_type(
        pyvizier.MetricType.OBJECTIVE
    )
    self._metric_name = objective_metrics.item().name

    self._problem_statement = copy.deepcopy(problem_statement)
    self._converter = converters.TrialToArrayConverter.from_study_config(
        study_config=self._problem_statement,
        scale=False,
        flip_sign_for_minimization_metrics=False,
    )

  def problem_statement(self) -> pyvizier.ProblemStatement:
    return copy.deepcopy(self._problem_statement)

  def evaluate(self, suggestions: Sequence[pyvizier.Trial]):
    # Features has shape (num_trials, num_features).
    features = self._converter.to_features(suggestions)
    for idx, suggestion in enumerate(suggestions):
      val = self.impl(features[idx])
      if math.isfinite(val):
        suggestion.complete(
            pyvizier.Measurement(metrics={self._metric_name: val})
        )
      else:
        self.problem_statement().search_space.assert_contains(
            suggestions[idx].parameters
        )
        suggestion.complete(
            pyvizier.Measurement(),
            infeasibility_reason='Objective value is not finite: %f' % val,
        )

  def __repr__(self) -> str:
    return f'NumpyExperimenter {{name: {self._impl_name}}}'



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Scheduled GP-UCB-PE for budget-aware optimization.
"""

import dataclasses as dc
from typing import Callable
import attrs
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import gp_ucb_pe
from vizier._src.algorithms.designers import scheduled_designer


@attrs.define(kw_only=True)
class ScheduledGPUCBPEFactory:
  """Scheduled GP-UCB-PE factory."""

  _gp_ucb_pe_factory: Callable[
      [vz.ProblemStatement], gp_ucb_pe.VizierGPUCBPEBandit
  ]
  _expected_total_num_trials: int
  # ucb_coefficient
  _init_ucb_coefficient: float
  _final_ucb_coefficient: float
  _decay_ucb_coefficient: float

  # explore_region_ucb_coefficient
  _init_explore_region_ucb_coefficient: float
  _final_explore_region_ucb_coefficient: float
  _decay_explore_region_ucb_coefficient: float

  # ucb_overwrite_probability
  _init_ucb_overwrite_probability: float
  _final_ucb_overwrite_probability: float
  _decay_ucb_overwrite_probability: float

  def __call__(
      self,
      problem: vz.ProblemStatement,
  ) -> scheduled_designer.ScheduledDesigner:
    """Creates a scheduled GP-Bandit designer."""

    def _gp_ucb_pe_state_updater(designer, params):
      # Create a new copy of the config and replace the ucb coefficient value.
      updated_config = dc.replace(
          designer._config,  # pylint: disable=protected-access
          ucb_coefficient=params['ucb_coefficient'],
          explore_region_ucb_coefficient=params[
              'explore_region_ucb_coefficient'
          ],
          ucb_overwrite_probability=params['ucb_overwrite_probability'],
      )
      # Update the designer config. As the GP_UCB_PE state is recomputed every
      # suggest() call based on the config, this effectively updates the state.
      # TODO: Instead of changing internal state, use public method.
      designer._config = updated_config  # pylint: disable=protected-access

    ucb_coefficient_param = scheduled_designer.ExponentialScheduledParam(
        init_value=self._init_ucb_coefficient,
        final_value=self._final_ucb_coefficient,
        rate=self._decay_ucb_coefficient,
    )

    explore_region_ucb_coefficient_param = (
        scheduled_designer.ExponentialScheduledParam(
            init_value=self._init_explore_region_ucb_coefficient,
            final_value=self._final_explore_region_ucb_coefficient,
            rate=self._decay_explore_region_ucb_coefficient,
        )
    )

    ucb_overwrite_probability_param = (
        scheduled_designer.ExponentialScheduledParam(
            init_value=self._init_ucb_overwrite_probability,
            final_value=self._final_ucb_overwrite_probability,
            rate=self._decay_ucb_overwrite_probability,
        )
    )

    scheduled_params = {
        'ucb_coefficient': ucb_coefficient_param,
        'explore_region_ucb_coefficient': explore_region_ucb_coefficient_param,
        'ucb_overwrite_probability': ucb_overwrite_probability_param,
    }

    return scheduled_designer.ScheduledDesigner(
        problem,
        designer_factory=self._gp_ucb_pe_factory,
        designer_state_updater=_gp_ucb_pe_state_updater,
        scheduled_params=scheduled_params,
        expected_total_num_trials=self._expected_total_num_trials,
    )



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Policy supporters that keep data in RAM."""

import copy
import datetime
from typing import Iterable, List, Optional, Sequence
import uuid

from absl import logging
import attr
import numpy as np
from vizier import pyvizier as vz
from vizier._src.pythia import policy
from vizier._src.pythia import policy_supporter
from vizier.pyvizier import converters
from vizier.pyvizier import multimetric


# TODO: Keep the Pareto frontier trials.
@attr.s(frozen=False, init=True, slots=True)
class InRamPolicySupporter(policy_supporter.PolicySupporter):
  """Runs a fresh Study in RAM using a Policy.

  InRamPolicySupporter acts as a limited vizier service + client that runs in
  RAM. Trials can only be added and never removed.

  Example of using a policy to run a Study for 100 iterations, 1 trial each:
    runner = InRamPolicySupporter(my_study_config)
    policy = MyPolicy(runner)
    for _ in range(100):
      trials = runner.SuggestTrials(policy, count=1)
      if not trials:
        break
      for t in trials:
        t.complete(vz.Measurement(
            {'my_objective': my_objective(t)}, inplace=True))

  Attributes:
    study_config: Study config.
    study_guid: Unique identifier for the current study.
  """

  study_config: vz.ProblemStatement = attr.ib(
      init=True,
      validator=attr.validators.instance_of(vz.ProblemStatement),
      on_setattr=attr.setters.frozen,
  )
  study_guid: str = attr.ib(
      init=True,
      kw_only=True,
      default='',
      converter=str,
      on_setattr=attr.setters.frozen,
  )
  _trials: dict[int, vz.Trial] = attr.ib(
      init=False, factory=dict, on_setattr=attr.setters.frozen
  )
  # Dictionary of study_guid to ProblemAndTrials.
  prior_studies: dict[str, vz.ProblemAndTrials] = attr.ib(
      init=False, factory=dict
  )

  def __str__(self) -> str:
    return (
        f'InRamPolicySupporter(study_guid={self.study_guid},'
        f' num_trials={len(self.trials)})'
    )

  @property
  def trials(self) -> Sequence[vz.Trial]:
    return list(self._trials.values())

  def study_descriptor(self) -> vz.StudyDescriptor:
    return vz.StudyDescriptor(
        self.study_config,
        guid=self.study_guid,
        max_trial_id=max(self._trials.keys()) if self._trials else 0,
    )

  def GetStudyConfig(
      self, study_guid: Optional[str] = None
  ) -> vz.ProblemStatement:
    if study_guid is None or study_guid == self.study_guid:
      return self.study_config
    elif study_guid in self.prior_studies:
      return self.prior_studies[study_guid].problem
    else:
      raise KeyError(f'Study does not exist in InRamSupporter: {study_guid}')

  def GetTrials(
      self,
      *,
      study_guid: Optional[str] = None,
      trial_ids: Optional[Iterable[int]] = None,
      min_trial_id: Optional[int] = None,
      max_trial_id: Optional[int] = None,
      status_matches: Optional[vz.TrialStatus] = None,
      include_intermediate_measurements: bool = True,
  ) -> List[vz.Trial]:
    """Returns trials by reference to allow changing their status and attributes."""
    self.CheckCancelled('GetTrials')
    if study_guid is not None and study_guid != self.study_guid:
      if study_guid not in self.prior_studies:
        raise KeyError(f'Study Guid does not exist {study_guid}')

      candidate_trials = self.prior_studies[study_guid].trials
    else:
      candidate_trials = self.trials

    trial_id_set = None
    if trial_ids is not None:
      trial_id_set = set(trial_ids)
    output: List[vz.Trial] = []
    for t in candidate_trials:
      if status_matches is not None and t.status != status_matches:
        continue
      if min_trial_id is not None and t.id < min_trial_id:
        continue
      if max_trial_id is not None and t.id > max_trial_id:
        continue
      if trial_id_set is not None and t.id not in trial_id_set:
        continue
      # NOTE: we ignore include_intermediate_measurements and always enclude
      # them.  That should be safe, and avoids a nasty conflict with the
      # pass-by-reference philosophy for Trials (you can't delete the
      # intermediate measurements without deleting them everywhere, and you
      # can't copy the trial without breaking desired reference connections).
      output.append(t)
    return output

  def CheckCancelled(self, note: Optional[str] = None) -> None:
    pass

  def TimeRemaining(self) -> datetime.timedelta:
    return datetime.timedelta(seconds=100.0)

  def _UpdateMetadata(self, delta: vz.MetadataDelta) -> None:
    """Assign metadata to trials."""
    for ns in delta.on_study.namespaces():
      self.study_config.metadata.abs_ns(ns).update(delta.on_study.abs_ns(ns))

    for tid, metadatum in delta.on_trials.items():
      if not tid > 0:
        raise ValueError(f'Bad Trial Id: {tid}')
      for ns in metadatum.namespaces():
        self._trials[tid].metadata.abs_ns(ns).update(metadatum.abs_ns(ns))

  # TODO: Return `count` trials for multi-objectives, when
  # `count` exceeds the size of the Pareto frontier.
  def GetBestTrials(self, *, count: Optional[int] = None) -> List[vz.Trial]:
    """Returns optimal trials.

    Single-objective study:
      * If `count` is unset, returns all tied top trials.
      * If `count` is set, returns top `count` trials, breaking ties
           arbitrarily.

    Multi-objective study:
      * If `count` is unset, returns all Pareto optimal trials.
      * If `count` is set, returns up to `count` of Pareto optimal trials that
          are arbitrarily selected.

    Args:
      count: If unset, returns Pareto optimal trials only. If set, returns the
        top "count" trials.

    Returns:
      Best trials.
    """
    if not self.study_config.metric_information.of_type(
        vz.MetricType.OBJECTIVE):
      raise ValueError('Requires at least one objective metric.')

    # Add safety warping and remove safety metrics from conversion.
    safety_checker = multimetric.SafetyChecker(
        self.study_config.metric_information
    )
    warped_trials = safety_checker.warp_unsafe_trials(
        copy.deepcopy(self.trials)
    )
    config_without_safe = copy.deepcopy(self.study_config)
    config_without_safe.metric_information = (
        self.study_config.metric_information.exclude_type(vz.MetricType.SAFETY)
    )
    converter = converters.TrialToArrayConverter.from_study_config(
        config_without_safe,
        flip_sign_for_minimization_metrics=True,
        dtype=np.float32,
    )

    if self.study_config.is_single_objective:
      # Single metric: Sort and take top N.
      count = count or 1  # Defaults to 1.
      labels = converter.to_labels(warped_trials).squeeze()
      sorted_idx = np.argsort(-labels)  # np.argsort sorts in ascending order.
      return list(np.asarray(self.trials)[sorted_idx[:count]])
    else:
      algorithm = multimetric.FastParetoOptimalAlgorithm()
      is_optimal = algorithm.is_pareto_optimal(
          points=converter.to_labels(warped_trials)
      )
      return list(np.asarray(self.trials)[is_optimal][:count])

  def SetPriorStudy(
      self, study: vz.ProblemAndTrials, study_guid: Optional[str] = None
  ) -> str:
    if study_guid is None:
      # Assign study_guid using unique identifier.
      study_guid = f'prior_{uuid.uuid1()}'
      if study_guid in self.prior_studies:
        raise RuntimeError(f'Cannot set unique id: {study_guid}')

    if study_guid in self.prior_studies:
      logging.warning('Prior study already exists with guid %s ', study_guid)

    self.prior_studies[study_guid] = study
    return study_guid

  def AddTrials(self, trials: Sequence[vz.Trial]) -> None:
    """Assigns ids to the trials and add them to the supporter (by reference).

    New IDs are always assigned in increasing order from the max id unless:
      * If an incoming Trial has an ID that matches an ACTIVE Trial, the ACTIVE
    Trial is replaced and the COMPLETED Trial keeps its ID.
      * If an incoming Trial has an ID that matches an COMPLETE Trial, no
    updates are done and there is a warning.

    Args:
      trials: Incoming Trials to add.
    """
    existing_trial_ids = self._trials.keys()
    next_trial_id = max(existing_trial_ids) + 1 if existing_trial_ids else 1
    for trial in trials:
      if trial.id in existing_trial_ids:
        if self._trials[trial.id].status == vz.TrialStatus.ACTIVE:
          self._trials[trial.id] = trial
        elif self._trials[trial.id].status == vz.TrialStatus.COMPLETED:
          logging.warning(
              'COMPLETED Trial %s cannot be overwritten by %s',
              self._trials[trial.id],
              trial,
          )
        continue

      trial.id = next_trial_id
      self._trials[next_trial_id] = trial
      next_trial_id += 1

  def AddSuggestions(
      self, suggestions: Iterable[vz.TrialSuggestion]) -> Sequence[vz.Trial]:
    """Assigns ids to suggestions and add them to the study."""
    trials = []
    for suggestion in suggestions:
      # Assign temporary ids, which will be overwritten by AddTrials() method.
      trials.append(suggestion.to_trial(0))
    self.AddTrials(trials)
    return trials

  def SuggestTrials(self, algorithm: policy.Policy,
                    count: int) -> Sequence[vz.Trial]:
    """Suggest and add new trials."""

    decisions = algorithm.suggest(
        policy.SuggestRequest(
            study_descriptor=self.study_descriptor(), count=count))
    self._UpdateMetadata(decisions.metadata)
    return self.AddSuggestions([
        vz.TrialSuggestion(d.parameters, metadata=d.metadata)
        for d in decisions.suggestions
    ])



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Common utility functions for COMBO benchmarks."""
# pylint:disable = missing-function-docstring
import itertools
from typing import Tuple
import numpy as np


def spin_covariance(interaction: Tuple[np.ndarray, np.ndarray],
                    grid_shape: Tuple[int, int]) -> Tuple[np.ndarray, float]:
  horizontal_interaction, vertical_interaction = interaction
  n_vars = horizontal_interaction.shape[0] * vertical_interaction.shape[1]
  spin_cfgs = np.array(list(itertools.product(*([[-1, 1]] * n_vars))))
  density = np.zeros(spin_cfgs.shape[0])
  for i in range(spin_cfgs.shape[0]):
    spin_cfg = spin_cfgs[i].reshape(grid_shape)
    h_comp = spin_cfg[:, :-1] * horizontal_interaction * spin_cfg[:, 1:] * 2
    v_comp = spin_cfg[:-1] * vertical_interaction * spin_cfg[1:] * 2
    log_interaction_energy = np.sum(h_comp) + np.sum(v_comp)
    density[i] = np.exp(log_interaction_energy)
  interaction_partition = np.sum(density)
  density = density / interaction_partition

  covariance = spin_cfgs.T.dot(spin_cfgs * density.reshape((-1, 1)))
  return covariance, interaction_partition


def partition(interaction: Tuple[np.ndarray, np.ndarray],
              grid_shape: Tuple[int, int]) -> float:
  horizontal_interaction, vertical_interaction = interaction
  n_vars = horizontal_interaction.shape[0] * vertical_interaction.shape[1]
  spin_cfgs = np.array(list(itertools.product(*([[-1, 1]] * n_vars))))
  interaction_partition = 0.0
  for i in range(spin_cfgs.shape[0]):
    spin_cfg = spin_cfgs[i].reshape(grid_shape)
    h_comp = spin_cfg[:, :-1] * horizontal_interaction * spin_cfg[:, 1:] * 2
    v_comp = spin_cfg[:-1] * vertical_interaction * spin_cfg[1:] * 2
    log_interaction_energy = np.sum(h_comp) + np.sum(v_comp)
    interaction_partition += np.exp(log_interaction_energy)

  return interaction_partition


def log_partition(interaction: Tuple[np.ndarray, np.ndarray],
                  grid_shape: Tuple[int, int]) -> float:
  horizontal_interaction, vertical_interaction = interaction
  n_vars = horizontal_interaction.shape[0] * vertical_interaction.shape[1]
  spin_cfgs = np.array(list(itertools.product(*([[-1, 1]] * n_vars))))
  log_interaction_energy_list = []
  for i in range(spin_cfgs.shape[0]):
    spin_cfg = spin_cfgs[i].reshape(grid_shape)
    h_comp = spin_cfg[:, :-1] * horizontal_interaction * spin_cfg[:, 1:] * 2
    v_comp = spin_cfg[:-1] * vertical_interaction * spin_cfg[1:] * 2
    log_interaction_energy = np.sum(h_comp) + np.sum(v_comp)
    log_interaction_energy_list.append(log_interaction_energy)

  log_interaction_energy_list = np.array(log_interaction_energy_list)
  max_log_interaction_energy = np.max(log_interaction_energy_list)
  interaction_partition = np.sum(
      np.exp(log_interaction_energy_list - max_log_interaction_energy))

  return np.log(interaction_partition) + max_log_interaction_energy


def generate_ising_interaction(
    grid_h: int,
    grid_w: int,
    random_seed=None) -> Tuple[np.ndarray, np.ndarray]:
  np.random.seed(random_seed)
  horizontal_interaction = (
      (np.random.randint(0, 2,
                         (grid_h * (grid_w - 1),)) * 2 - 1).astype(float) *
      (np.random.rand(grid_h * (grid_w - 1)) * (5 - 0.05) + 0.05)).reshape(
          grid_h, grid_w - 1)
  vertical_interaction = ((np.random.randint(0, 2, (
      (grid_h - 1) * grid_w,)) * 2 - 1).astype(float) * (np.random.rand(
          (grid_h - 1) * grid_w) * (5 - 0.05) + 0.05)).reshape(
              grid_h - 1, grid_w)
  return horizontal_interaction, vertical_interaction


def ising_dense(ising_grid_h: int, interaction_original: Tuple[np.ndarray,
                                                               np.ndarray],
                interaction_sparsified: Tuple[np.ndarray, np.ndarray],
                covariance: np.ndarray, log_partition_original: float,
                log_partition_new: float) -> float:
  diff_horizontal = interaction_original[0] - interaction_sparsified[0]
  diff_vertical = interaction_original[1] - interaction_sparsified[1]

  kld = 0
  n_spin = covariance.shape[0]
  for i in range(n_spin):
    i_h, i_v = int(i / ising_grid_h), int(i % ising_grid_h)
    for j in range(i, n_spin):
      j_h, j_v = int(j / ising_grid_h), int(j % ising_grid_h)
      if i_h == j_h and abs(i_v - j_v) == 1:
        kld += diff_horizontal[i_h, min(i_v, j_v)] * covariance[i, j]
      elif abs(i_h - j_h) == 1 and i_v == j_v:
        kld += diff_vertical[min(i_h, j_h), i_v] * covariance[i, j]

  return kld * 2 + log_partition_new - log_partition_original



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Collection of multitask GP models."""

import enum
import functools
from typing import Any, Generator, Optional, Union

from flax import struct
import jax
from jax import numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp
from vizier._src.jax import stochastic_process_model as sp
from vizier._src.jax import types
from vizier._src.jax.models import continuous_only_kernel
from vizier._src.jax.models import mask_features


tfb = tfp.bijectors
tfd = tfp.distributions
tfde = tfp.experimental.distributions
tfpk = tfp.math.psd_kernels
tfpke = tfp.experimental.psd_kernels


class MultiTaskType(enum.Enum):
  """The form of the MultiTask GP."""

  # Use a TFP MultiTask GP with an Independent kernel. A base (prior) kernel
  # matrix is shared across all tasks, and the overall kernel matrix is the
  # Kronecker product of the base kernel matrix and the identity.
  INDEPENDENT = 'independent'

  # Use a TFP MultiTask GP with a Separable kernel. The prior kernel matrix
  # is the Kronecker product of a base kernel matrix and a task kernel matrix
  # (assumed to be full-covariance).
  SEPARABLE_NORMAL_TASK_KERNEL_PRIOR = 'separable_normal_task_kernel_prior'

  # Use an LKJ-distributed prior for the (full-covariance) Separable task
  # kernel.
  SEPARABLE_LKJ_TASK_KERNEL_PRIOR = 'separable_lkj_task_kernel_prior'

  # Use a diagonal matrix for the Separable task kernel.
  SEPARABLE_DIAG_TASK_KERNEL_PRIOR = 'separable_diag_task_kernel_prior'


@struct.dataclass
class VizierMultitaskGaussianProcess(
    sp.ModelCoroutine[Union[tfd.GaussianProcess, tfde.MultiTaskGaussianProcess]]
):
  """Multitask GP model using priors from Vizier's tuned GP."""

  _feature_dim: types.ContinuousAndCategorical[int] = struct.field(
      pytree_node=False
  )
  _num_tasks: int = struct.field(pytree_node=False)
  _multitask_type: MultiTaskType = struct.field(
      default=MultiTaskType.INDEPENDENT, kw_only=True, pytree_node=False
  )
  _use_retrying_cholesky: bool = struct.field(
      default=True, kw_only=True, pytree_node=False
  )
  _boundary_epsilon: float = struct.field(default=1e-12, kw_only=True)

  def _log_uniform_init(
      self,
      low: Union[float, np.floating],
      high: Union[float, np.floating],
      shape: tuple[int, ...] = tuple(),
  ) -> sp.InitFn:
    r"""Take log-uniform sample in the constraint and map it back to \R.

    Args:
      low: Parameter lower bound.
      high: Parameter upper bound.
      shape: Returned array has this shape. Each entry in the returned array is
        an i.i.d sample.

    Returns:
      Randomly sampled array.
    """

    def sample(key: Any) -> jnp.ndarray:
      unif = jax.random.uniform(key, shape, dtype=jnp.float64)
      return jnp.exp(unif * jnp.log(high / low) + jnp.log(low))

    return sample

  def _build_task_kernel_scale_linop(
      self,
  ) -> Generator[
      sp.ModelParameter, jax.Array, tfp.tf2jax.linalg.LinearOperator
  ]:
    if self._multitask_type == MultiTaskType.SEPARABLE_DIAG_TASK_KERNEL_PRIOR:
      correlation_diag = yield sp.ModelParameter.from_prior(
          tfd.Sample(
              tfd.Uniform(low=jnp.float64(1e-6), high=1.0),
              sample_shape=self._num_tasks,
              name='correlation_diag',
          ),
          constraint=sp.Constraint(
              bounds=(1e-6, 1.0),
              bijector=tfb.Sigmoid(low=jnp.float64(1e-6), high=1.0),
          ),
      )
      task_kernel_scale_linop = tfp.tf2jax.linalg.LinearOperatorDiag(
          correlation_diag
      )
    elif self._multitask_type == MultiTaskType.SEPARABLE_LKJ_TASK_KERNEL_PRIOR:
      # Generate parameters for the Cholesky of the task kernel matrix,
      # which accounts for correlations between tasks.
      num_task_kernel_entries = tfb.CorrelationCholesky().inverse_event_shape(
          [self._num_tasks, self._num_tasks]
      )
      correlation_cholesky_vec = yield sp.ModelParameter(
          init_fn=lambda key: tfd.Sample(  # pylint: disable=g-long-lambda
              tfd.Normal(jnp.float64(0.0), 1.0), num_task_kernel_entries
          ).sample(seed=key),
          # Use `jnp.copy` to prevent tracers leaking from bijector cache.
          regularizer=lambda x: -tfd.CholeskyLKJ(  # pylint: disable=g-long-lambda
              dimension=self._num_tasks, concentration=1.0
          ).log_prob(tfb.CorrelationCholesky()(jnp.copy(x))),
          name='task_kernel_correlation_cholesky_vec',
      )

      task_kernel_correlation_cholesky = tfb.CorrelationCholesky()(
          jnp.copy(correlation_cholesky_vec)
      )

      task_kernel_scale_vec = yield sp.ModelParameter(
          init_fn=functools.partial(
              jax.random.uniform,
              shape=(self._num_tasks,),
              dtype=jnp.float64,
              minval=1e-6,
              maxval=1.0,
          ),
          constraint=sp.Constraint(
              bounds=(1e-6, 1.0),
              bijector=tfb.Sigmoid(low=jnp.float64(1e-6), high=1.0),
          ),
          name='task_kernel_sqrt_diagonal',
      )
      task_kernel_cholesky = (
          task_kernel_correlation_cholesky
          * task_kernel_scale_vec[:, jnp.newaxis]
      )

      # Build the `LinearOperator` object representing the task kernel matrix,
      # to parameterize the Separable kernel.
      task_kernel_scale_linop = tfp.tf2jax.linalg.LinearOperatorLowerTriangular(
          task_kernel_cholesky
      )
    elif (
        self._multitask_type == MultiTaskType.SEPARABLE_NORMAL_TASK_KERNEL_PRIOR
    ):
      # Generate parameters for the Cholesky of the task kernel matrix;
      # accounts for correlations between tasks. The task kernel matrix must
      # be positive definite, so we construct it via a Cholesky factor.
      # Define the prior of the kernel task matrix to be centered at the
      # identity.
      prior_mean = jnp.eye(self._num_tasks, dtype=jnp.float64)
      prior_mean_vec = tfb.FillTriangular().inverse(prior_mean)
      prior_mean_batched = jnp.broadcast_to(
          prior_mean_vec, prior_mean_vec.shape
      )

      task_kernel_cholesky_entries = yield sp.ModelParameter.from_prior(
          tfd.Independent(
              tfd.Normal(prior_mean_batched, 1.0),
              reinterpreted_batch_ndims=1,
              name='task_kernel_cholesky_entries',
          )
      )

      # Apply a bijector to pack the task kernel entries into a lower
      # triangular matrix and ensure the diagonal is positive.
      task_kernel_bijector = tfb.Chain([
          tfb.TransformDiagonal(
              tfb.Chain([tfb.Shift(jnp.float64(1e-6)), tfb.Softplus()])
          ),
          tfb.FillTriangular(),
      ])
      task_kernel_cholesky = task_kernel_bijector(
          jnp.copy(task_kernel_cholesky_entries)
      )

      # Build the `LinearOperator` object representing the task kernel
      # matrix, to parameterize the Separable kernel.
      task_kernel_scale_linop = tfp.tf2jax.linalg.LinearOperatorLowerTriangular(
          task_kernel_cholesky
      )
    else:
      raise ValueError(f'Unsupported multitask type: {self._multitask_type}')

    return task_kernel_scale_linop

  def __call__(
      self, inputs: Optional[types.ModelInput] = None
  ) -> Generator[
      sp.ModelParameter,
      jax.Array,
      Union[tfd.GaussianProcess, tfde.MultiTaskGaussianProcess],
  ]:
    """Creates a generator.

    Args:
      inputs: Floating array of dimension (num_examples, _feature_dim).

    Yields:
      Multitask GP hyperparameters.
    """

    eps = self._boundary_epsilon
    observation_noise_bounds = (np.float64(1e-10 - eps), 1.0 + eps)
    amplitude_bounds = (np.float64(1e-3 - eps), 10.0 + eps)
    continuous_ones = np.ones((self._feature_dim.continuous), dtype=np.float64)
    continuous_length_scale_bounds = (
        continuous_ones * (1e-2 - eps),
        continuous_ones * 1e2 + eps,
    )
    categorical_ones = np.ones(
        (self._feature_dim.categorical), dtype=np.float64
    )
    categorical_length_scale_bounds = (
        categorical_ones * (1e-2 - eps),
        categorical_ones * 1e2 + eps,
    )

    signal_variance = yield sp.ModelParameter(
        init_fn=self._log_uniform_init(*amplitude_bounds),
        constraint=sp.Constraint(
            amplitude_bounds,
            tfb.SoftClip(*amplitude_bounds, hinge_softness=1e-2),
        ),
        regularizer=lambda x: 0.01 * jnp.log(x / 0.039) ** 2,
        name='signal_variance',
    )
    kernel = tfpk.MaternFiveHalves(amplitude=jnp.sqrt(signal_variance))
    continuous_length_scale_squared = yield sp.ModelParameter(
        init_fn=self._log_uniform_init(
            *continuous_length_scale_bounds,
            shape=(self._feature_dim.continuous,),
        ),
        constraint=sp.Constraint(
            continuous_length_scale_bounds,
            tfb.SoftClip(*continuous_length_scale_bounds, hinge_softness=1e-2),
        ),
        regularizer=lambda x: jnp.sum(0.01 * jnp.log(x / 0.5) ** 2),
        name='continuous_length_scale_squared',
    )
    categorical_length_scale_squared = yield sp.ModelParameter(
        init_fn=self._log_uniform_init(
            *categorical_length_scale_bounds,
            shape=(self._feature_dim.categorical,),
        ),
        constraint=sp.Constraint(
            categorical_length_scale_bounds,
            tfb.SoftClip(*categorical_length_scale_bounds, hinge_softness=1e-2),
        ),
        regularizer=lambda x: jnp.sum(0.01 * jnp.log(x / 0.5) ** 2),
        name='categorical_length_scale_squared',
    )

    kernel = tfpke.FeatureScaledWithCategorical(
        kernel,
        scale_diag=tfpke.ContinuousAndCategoricalValues(
            jnp.sqrt(continuous_length_scale_squared),
            jnp.sqrt(categorical_length_scale_squared),
        ),
    )

    bias_amplitude = yield sp.ModelParameter.from_prior(
        tfd.Normal(
            jnp.zeros(shape=tuple(), dtype=jnp.float64),
            1.0,
            name='bias_amplitude',
        ),
    )
    slope_amplitude = yield sp.ModelParameter.from_prior(
        tfd.Normal(
            jnp.zeros(shape=tuple(), dtype=jnp.float64),
            1.0,
            name='slope_amplitude',
        ),
    )
    shift = yield sp.ModelParameter.from_prior(
        tfd.Normal(
            jnp.zeros(shape=tuple(), dtype=jnp.float64),
            1.0,
            name='shift',
        )
    )
    kernel = kernel + continuous_only_kernel.ContinuousOnly(
        tfpk.FeatureScaled(
            tfpk.Linear(
                slope_amplitude=slope_amplitude,
                bias_amplitude=bias_amplitude,
                shift=shift,
            ),
            scale_diag=jnp.sqrt(continuous_length_scale_squared),
        )
    )

    observation_noise_variance = yield sp.ModelParameter(
        init_fn=self._log_uniform_init(*observation_noise_bounds),
        constraint=sp.Constraint(
            observation_noise_bounds,
            tfb.SoftClip(*observation_noise_bounds, hinge_softness=1e-2),
        ),
        regularizer=lambda x: 0.01 * jnp.log(x / 0.0039) ** 2,
        name='observation_noise_variance',
    )

    cholesky_fn = None
    # When cholesky fails, increase jitters and retry.
    if self._use_retrying_cholesky:
      retrying_cholesky = functools.partial(
          tfp.experimental.distributions.marginal_fns.retrying_cholesky,
          jitter=np.float64(1e-4),
          max_iters=5,
      )
      cholesky_fn = lambda matrix: retrying_cholesky(matrix)[0]

    if inputs is not None:
      kernel = mask_features.MaskFeatures(
          kernel,
          dimension_is_missing=tfpke.ContinuousAndCategoricalValues(
              continuous=inputs.continuous.is_missing[1],
              categorical=inputs.categorical.is_missing[1],
          ),
      )

      inputs = tfpke.ContinuousAndCategoricalValues(
          inputs.continuous.padded_array,
          inputs.categorical.padded_array,
      )

    # Creates multitask kernel from single-task kernel. Defaults to INDEPENDENT
    # for multitask kernel.
    if self._multitask_type == MultiTaskType.INDEPENDENT:
      multitask_kernel = tfpke.Independent(self._num_tasks, kernel)
    else:
      task_kernel_scale_linop = yield from self._build_task_kernel_scale_linop()
      multitask_kernel = tfpke.Separable(
          self._num_tasks,
          base_kernel=kernel,
          task_kernel_scale_linop=task_kernel_scale_linop,
      )

    return tfde.MultiTaskGaussianProcess(
        multitask_kernel,
        index_points=inputs,
        observation_noise_variance=observation_noise_variance,
        cholesky_fn=cholesky_fn,
    )



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Abstractions and default converters."""

import abc
import copy
import dataclasses
import enum
from typing import Any, Callable, Collection, Dict, Iterator, List, Mapping, Optional, Sequence, Tuple, Type, Union

from absl import logging
import attr
import numpy as np
from vizier import pyvizier

# The study identifier for cross-study learning must be stored in
# Trial.Metadata and StudyConfig.Metadata with this key.
# TODO: Use metadata namespace instead.
STUDY_ID_FIELD = 'metalearn_study_id'


class NumpyArraySpecType(enum.Enum):
  """Type information for NumpyArraySpec.

  CONTINUOUS: Continuous parameter
  DISCRETE: Discrete/integer/categorical parameter
  ONEHOT_EMBEDDING: One-hot embedding of DISCRETE.
  OBJECT: Object type that can hold any value.
  """

  CONTINUOUS = 'CONTINUOUS'
  DISCRETE = 'DISCRETE'
  ONEHOT_EMBEDDING = 'ONEHOT_EMBEDDING'
  OBJECT = 'OBJECT'

  @classmethod
  def default_factory(
      cls, pc: pyvizier.ParameterConfig
  ) -> 'NumpyArraySpecType':
    """SpecType when encoding discretes as integer indices."""
    if pc.type == pyvizier.ParameterType.DOUBLE:
      return NumpyArraySpecType.CONTINUOUS
    elif pc.type in (
        pyvizier.ParameterType.DISCRETE,
        pyvizier.ParameterType.CATEGORICAL,
        pyvizier.ParameterType.INTEGER,
    ):
      return NumpyArraySpecType.DISCRETE
    elif pc.type == pyvizier.ParameterType.CUSTOM:
      return NumpyArraySpecType.OBJECT
    raise ValueError(f'Unknown type {pc.type}')

  @classmethod
  def embedding_factory(
      cls, pc: pyvizier.ParameterConfig
  ) -> 'NumpyArraySpecType':
    """SpecType when encoding discretes as onehot embedding."""
    if pc.type == pyvizier.ParameterType.DOUBLE:
      return NumpyArraySpecType.CONTINUOUS
    elif pc.type in (
        pyvizier.ParameterType.DISCRETE,
        pyvizier.ParameterType.CATEGORICAL,
        pyvizier.ParameterType.INTEGER,
    ):
      return NumpyArraySpecType.ONEHOT_EMBEDDING
    raise ValueError(f'Unknown type {pc.type}')


@attr.define(frozen=True, auto_attribs=True, eq=True, hash=True)
class NumpyArraySpec:
  """Encodes what a feature array represents.

  This class can be seen as a counterpart of Vizier ParameterConfig.
  Vizier ParameterConfigs describe Trial parameters. NumpyArraySpec describes
  numpy arrays returned from trial-to-numpy converters.

  This class is similar to `BoundedTensorSpec` in tf agents, except it carries
  extra information specific to vizier.

  If `type` is `DOUBLE`, then `dtype` is a floating type, and bounds are
  floating numbers. num_dimensions is always 1, and num_oovs is zero.

  If 'type' is `DISCRETE`, then `dtype` is an integer type, and bounds are
  integers. num_dimensions is always 1. Suppose `bounds=(x,y)`. Then integers
  x to-and-including (y-num_oovs) correspond to valid parameter values. The rest
  represent out-of-vocabulary values. For example, an integer parameter in
  the range 1 <= x <= 3 can be represented by a DISCRETE NumpyArraySpec with
  bounds=(1,4) and oov=1.

  If 'type' is `ONEHOT_EMBEDDING`, then `dtype` is a floating type, and bounds
  are floating numbers. In this case, the output array can have multiple
  columns.

  Attributes:
    type: Underlying type of the Vizier parameter corresponding to the array.
    dtype: The numpy array's type.
    bounds: Always inclusive in both directions.
    num_dimensions: Corresponds to shape[-1] of the numpy array. When `type` is
      `ONEHOT_EMBEDDING`, the first X dimensions correspond to valid parameter
      values. The other dimensions correspond to out-of-vocabulary values.
    name: Parameter name.
    num_oovs: Number of out-of-vocabulary items, for non-continuous type.
    scale: Scaling of the values.
  """

  type: NumpyArraySpecType = attr.field(
      validator=attr.validators.instance_of(NumpyArraySpecType)
  )
  dtype: np.dtype = attr.field(
      converter=np.dtype,
      validator=attr.validators.in_(
          [np.float32, np.int32, np.float64, np.int64, np.object_]
      ),
  )
  bounds: Union[Tuple[float, float], Tuple[int, int]]
  num_dimensions: int = attr.field(validator=attr.validators.instance_of(int))
  name: str = attr.field(validator=attr.validators.instance_of(str))
  num_oovs: int = attr.field(validator=attr.validators.instance_of(int))
  scale: Optional[pyvizier.ScaleType] = attr.field(
      default=None,
      validator=attr.validators.optional(
          attr.validators.instance_of(pyvizier.ScaleType)
      ),
  )

  def __attrs_post_init__(self):
    object.__setattr__(
        self, 'bounds', tuple(np.array(self.bounds, dtype=self.dtype))
    )

  @classmethod
  def from_parameter_config(
      cls,
      pc: pyvizier.ParameterConfig,
      type_factory: Callable[
          [pyvizier.ParameterConfig], NumpyArraySpecType
      ] = NumpyArraySpecType.default_factory,
      floating_dtype: Union[np.dtype, Type[np.floating]] = np.float32,
      int_dtype: Union[np.dtype, Type[np.integer]] = np.int32,
      *,
      pad_oovs: bool = True,
  ) -> 'NumpyArraySpec':
    """Factory function.

    Args:
      pc:
      type_factory: NumpyArraySpecType has `default_factory` and
        `embedding_factory`. The difference is in how they handle non-continuous
        parameters.
      floating_dtype: Dtype of the floating outputs.
      int_dtype: Dtype of the integer outputs.
      pad_oovs: If True, pad the out-of-vocabulary dimensions to onehot
        embedding.

    Returns:
      NumpyArraySpec.
    """
    the_type = type_factory(pc)
    if the_type == NumpyArraySpecType.CONTINUOUS:
      return NumpyArraySpec(
          the_type,
          np.dtype(floating_dtype),
          bounds=pc.bounds,
          num_dimensions=1,
          scale=pc.scale_type,
          name=pc.name,
          num_oovs=0,
      )
    elif the_type == NumpyArraySpecType.DISCRETE:
      return NumpyArraySpec(
          the_type,
          np.dtype(int_dtype),
          bounds=(0, len(pc.feasible_values)),
          num_dimensions=1,
          name=pc.name,
          num_oovs=1 if pad_oovs else 0,
      )
    elif the_type == NumpyArraySpecType.ONEHOT_EMBEDDING:
      return NumpyArraySpec(
          the_type,
          np.dtype(floating_dtype),
          bounds=(0.0, 1.0),
          num_dimensions=len(pc.feasible_values) + 1,
          name=pc.name,
          num_oovs=1 if pad_oovs else 0,
      )
    elif the_type == NumpyArraySpecType.OBJECT:
      return NumpyArraySpec(
          the_type,
          dtype=np.object_,
          bounds=(0, 0),
          num_dimensions=0,
          name=pc.name,
          num_oovs=0,
      )
    raise ValueError(f'Unknown type {type}')


def dict_to_array(array_dict: Mapping[Any, np.ndarray]) -> np.ndarray:
  r"""Converts a dict of (..., D_i) arrays to a (..., \sum_i D_i) array."""
  return np.concatenate(list(array_dict.values()), axis=-1)


class DictOf2DArrays(Mapping[str, np.ndarray]):
  """Dictionary of string to 2D arrays.

  All arrays share the first dimension, which is at a high level, the number of
  objects that this dictionary corresponds to.

  Attributes:
    size: Array's shape[0].
  """

  def __init__(self, d: Mapping[str, np.ndarray]):
    self._d = d
    shape = None
    for k, v in self.items():
      if shape is None:
        shape = v.shape
        if len(shape) != 2:
          raise ValueError(
              f'{k} has shape {v.shape} which is not length 2.'
              'DictOf2DArrays only supports 2D numpy arrays.'
          )
      if shape[0] != v.shape[0]:
        raise ValueError(
            f'{k} has shape {v.shape} which is not equal to {shape}.'
        )
    self._size = shape[0]

  def __getitem__(self, key: str) -> np.ndarray:
    return self._d[key]

  def __iter__(self) -> Iterator[str]:
    return iter(self._d)

  def __len__(self) -> int:
    return len(self._d)

  def __add__(self, other: 'DictOf2DArrays') -> 'DictOf2DArrays':
    if not isinstance(other, DictOf2DArrays):
      raise ValueError('You can add DictOf2DArrays only.')
    if len(self) != len(other):
      # We don't check the keys because it's too slow.
      raise ValueError('Two arrays have different length!')
    return DictOf2DArrays(
        {k: np.concatenate([self[k], other[k]], axis=0) for k in self}
    )

  @property
  def size(self) -> int:
    return self._size

  def asarray(self) -> np.ndarray:
    return dict_to_array(self._d)

  def dict_like(self, array: np.ndarray) -> 'DictOf2DArrays':
    """[Experimental] Converts an array into a dict with the same keys as this.

    This function acts like an inverse of `asarray()`, i.e. it satisfies
      `self.dict_like(self.asarray()) == self`.

    Example:
      d = DictOf2DArrays({'p1': [[1], [2], [3]], 'p2': [[4], [5], [6]]})
      d.dict_like([[1, 2], [3, 4]]) == {'p1': [[1], [3]], 'p2': [[2],[4]]}

    Args:
      array:

    Returns:
      DictOf2DArrays with the same shape spec as this.
    """
    begin = 0
    new_dict = dict()
    for k, v in self.items():
      end = begin + v.shape[1]
      new_dict[k] = array[:, begin:end].astype(v.dtype)
      begin = end

    return DictOf2DArrays(new_dict)


class TrialToNumpyDict(abc.ABC):
  """Parses a sequence of Trials to a dict keyed by parameter and metric names.

  Design note:
    A typical Keras pipeline consists of:
      1. Load data into arrays.
      2. Call Model.build() to initialize a model for the loaded data shape.
      3. Call Model.fit() to train the model.
      4. Call Model.__call__() to predict with the model.
    This abstraction allows us to switch implementations for the step number 1.

  This class consists of `to_xy()` function and properties that describe
  the returned values of `to_xy()`. Subclasses are free to have extra functions
  for convenience.
  """

  @abc.abstractmethod
  def to_xy(
      self, trials: Sequence[pyvizier.Trial]
  ) -> Tuple[Dict[str, np.ndarray], Dict[str, np.ndarray]]:
    """Extracts features-labels pairs that can be used for fitting a model."""
    pass

  @property
  @abc.abstractmethod
  def features_shape(self) -> Dict[str, Sequence[Union[int, None]]]:
    """Describes the shape of the first element returned from to_xy().

    Returns:
      Dict of feature names to shapes. `None` means that the shape depends
      on the input "trials".
    # TODO: Add compute_feature_shape(trials) that returns exact
    shapes.
    """
    pass

  @property
  @abc.abstractmethod
  def output_specs(self) -> Dict[str, NumpyArraySpec]:
    """Describes the semantics of the first element returned from to_xy().

    See NumpyArraySpec for more details.

    Returns:
      Dict of feature names to NumpyArraySpecs.
    """
    pass

  @property
  @abc.abstractmethod
  def labels_shape(self) -> Dict[str, Any]:
    """Describes the shape of the second element returned from to_xy().

    Returns:
      Dict of label names to shapes. `None` means that the shape depends
      on the input "trials".
    """
    pass

  @property
  @abc.abstractmethod
  def metric_information(self) -> Dict[str, pyvizier.MetricInformation]:
    """Describes the semantics of the second element returned from to_xy().

    See ModelOutputConverter for more details.
    Returns:
      Dict of label names to MetricInformation.
    """


class ModelInputConverter(metaclass=abc.ABCMeta):
  """Interface for extracting inputs to the model."""

  @abc.abstractmethod
  def convert(self, trials: Sequence[pyvizier.TrialSuggestion]) -> np.ndarray:
    """Returns an array of shape (number of trials, feature dimension).

    Args:
      trials:

    Returns:
      Returns an array of shape (number of trials, feature dimension).
      Subclasses must use a fixed feature dimension. In particular, it should
      be a constant function of the input trials.
    """

  @property
  @abc.abstractmethod
  def output_spec(self) -> NumpyArraySpec:
    """Provides specification of the output from this converter."""

  @property
  @abc.abstractmethod
  def parameter_config(self):
    """Original ParameterConfig that this converter acts on."""

  @abc.abstractmethod
  def to_parameter_values(
      self, array: np.ndarray
  ) -> List[Optional[pyvizier.ParameterValue]]:
    """Convert and clip to the nearest feasible parameter values.

    NOTE: value is automatically truncated to lie inside the search space.
      This method should only be called to convert suggestions.

    Args:
      array: has shape (number of trials, feature dimension)

    Returns:
      List of (ParameterValue or None).  A list entry is None when this
      converter's parameter doesn't exist in the Trial.  (This is common
      when parameters are conditional.)
    """


@dataclasses.dataclass
class ModelInputArrayBijector:
  """Transformations on the numpy arrays generated by ModelInputConverter."""

  forward_fn: Callable[[np.ndarray], np.ndarray]
  backward_fn: Callable[[np.ndarray], np.ndarray]
  output_spec: NumpyArraySpec  # Spec after forward_fn is applied.

  @classmethod
  def identity(cls, spec: NumpyArraySpec) -> 'ModelInputArrayBijector':
    return cls(lambda x: x, lambda x: x, spec)

  @classmethod
  def scaler_from_spec(cls, spec: NumpyArraySpec) -> 'ModelInputArrayBijector':
    """For continuous specs, linearize and scale it to (0, 1) range."""
    low, high = spec.bounds
    if spec.type != NumpyArraySpecType.CONTINUOUS:
      return cls.identity(attr.evolve(spec, scale=None))
    if low == high:

      def backward_fn(y):
        return np.where(np.isfinite(y), y + low - 0.5, y)

      def forward_fn(y):
        return np.where(np.isfinite(y), y - low + 0.5, y)

      return cls(
          forward_fn,
          backward_fn,
          attr.evolve(spec, bounds=(0.5, 0.5), scale=None),
      )

    if spec.scale == pyvizier.ScaleType.LOG:
      if low < 0 or high < 0:
        raise ValueError(
            'Log scale requires both parameter boundaries to be positive,'
            f' though low bound is {low} and high bound is {high}.'
        )
      low, high = np.log(low), np.log(high)
      denom = (high - low) or 1.0
      if denom < 1e-6:
        logging.warning('Unusually small range detected for %s', spec)
      scale_fn = lambda x, low=low, denom=denom: (np.log(x) - low) / denom
      unscale_fn = lambda x, low=low, denom=denom: np.exp(x * denom + low)
    elif spec.scale == pyvizier.ScaleType.REVERSE_LOG:
      raw_sum = low + high
      low, high = np.log(low), np.log(high)
      denom = (high - low) or 1.0
      if denom < 1e-6:
        logging.warning('Unusually small range detected for %s', spec)

      def scale_fn(x, low=low, raw_sum=raw_sum, denom=denom):
        return 1.0 - (np.log(raw_sum - x) - low) / denom

      def unscale_fn(x, high=high, raw_sum=raw_sum):
        return raw_sum - np.exp(high - denom * x)

    else:
      if not (spec.scale == pyvizier.ScaleType.LINEAR or spec.scale is None):
        logging.warning('Unknown scale type %s. Applying LINEAR', spec.scale)
      denom = high - low
      if denom < 1e-6:
        logging.warning('Unusually small range detected for %s', spec)
      if denom == 1.0 and low == 0:
        return cls.identity(attr.evolve(spec, scale=None))
      scale_fn = lambda x, high=high, low=low: (x - low) / (high - low)
      unscale_fn = lambda x, high=high, low=low: x * (high - low) + low

    return cls(
        scale_fn, unscale_fn, attr.evolve(spec, bounds=(0.0, 1.0), scale=None)
    )

  @classmethod
  def onehot_embedder_from_spec(
      cls, spec: NumpyArraySpec, *, dtype=np.float32, pad_oovs: bool = True
  ) -> 'ModelInputArrayBijector':
    """Given a discrete spec, one-hot embeds it."""
    if spec.type != NumpyArraySpecType.DISCRETE:
      return cls.identity(spec)

    num_oovs = 1 if pad_oovs else 0
    output_spec = NumpyArraySpec(
        NumpyArraySpecType.ONEHOT_EMBEDDING,
        dtype,
        bounds=(0.0, 1.0),
        num_dimensions=int(spec.bounds[1] - spec.bounds[0] + num_oovs),
        name=spec.name,
        num_oovs=num_oovs,
        scale=None,
    )

    def embed_fn(x: np.ndarray, output_spec=output_spec):
      """x is integer array of [N, 1]."""
      return np.eye(output_spec.num_dimensions, dtype=output_spec.dtype)[
          x.flatten()
      ]

    def unembed_fn(x: np.ndarray, spec=spec, output_spec=output_spec):
      return np.argmax(
          x[:, : output_spec.num_dimensions - output_spec.num_oovs], axis=1
      ).astype(spec.dtype)

    return cls(embed_fn, unembed_fn, output_spec)


def _create_default_getter(
    pconfig: pyvizier.ParameterConfig,
) -> Callable[[pyvizier.TrialSuggestion], Any]:
  """Create a default getter for the given parameter config."""

  def getter(trial, pconfig=pconfig):
    if pconfig.name not in trial.parameters:
      return None

    pvalue = trial.parameters[pconfig.name]
    if pconfig.type == pyvizier.ParameterType.DOUBLE:
      return pvalue.as_float
    elif pconfig.type == pyvizier.ParameterType.DISCRETE:
      return pvalue.as_float
    elif pconfig.type == pyvizier.ParameterType.INTEGER:
      return pvalue.as_int
    else:
      return pvalue.as_str

  return getter


class DefaultModelInputConverter(ModelInputConverter):
  """Converts trials into a (None, 1) array corresponding to a parameter.

  If the parameter_config is continuous, values obtained from `getter()` are
  directly returned as floating numbers. Otherwise, this converter returns
  the index of the value obtained from `getter()` within
  `parameter_config.feasible_points` as int32.
  """

  def __init__(
      self,
      parameter_config: pyvizier.ParameterConfig,
      getter: Optional[Callable[[pyvizier.TrialSuggestion], Any]] = None,
      *,
      float_dtype: Union[np.dtype, Type[np.floating]] = np.float32,
      max_discrete_indices: int = 10,
      scale: bool = False,
      onehot_embed: bool = False,
      converts_to_parameter: bool = True,
      pad_oovs: bool = True,
      should_clip: bool = True,
  ):
    """Init.

    Given B trials, convert() always converts to (B, 1) array. The returned
    array may contain NaNs.

    Args:
      parameter_config:
      getter: See class pydoc. If the getter is not specified, the default
        getter looks up `parameter_config.name` inside `Trial.parameters`.
      float_dtype: floating precision to be used.
      max_discrete_indices: If the parameter config has more than this many
        number of DISCRETE/INTEGER feasible points, then the parameter config is
        continuified first.
      scale:
      onehot_embed:
      converts_to_parameter: If False, this converter does not correspond to an
        actual parameter in Vizier search space, and `to_parameter_value` always
        returns None
      pad_oovs: If True, pad the out-of-vocabulary dimensions to onehot
        embedding.
      should_clip: If True, clipping should be applied to parameter values.
    """
    self._converts_to_parameter = converts_to_parameter
    self._parameter_config = copy.deepcopy(parameter_config)
    if (
        parameter_config.type
        in (pyvizier.ParameterType.INTEGER, pyvizier.ParameterType.DISCRETE)
        and parameter_config.num_feasible_values > max_discrete_indices
    ):
      parameter_config = parameter_config.continuify()

    # TODO: Make the default getter raise an Error if they encounter an
    # out-of-vocabulary value but pad_oovs is False.
    self._getter = getter or _create_default_getter(parameter_config)
    # Getter spec can only have DISCRETE or CONTINUOUS types.
    self._getter_spec = NumpyArraySpec.from_parameter_config(
        parameter_config,
        NumpyArraySpecType.default_factory,
        floating_dtype=float_dtype,
    )

    # Optionally scale and onehot embed.
    spec = self._getter_spec
    self.scaler = (
        ModelInputArrayBijector.scaler_from_spec(spec)
        if scale
        else ModelInputArrayBijector.identity(spec)
    )
    spec = self.scaler.output_spec
    if onehot_embed:
      self.onehot_encoder = ModelInputArrayBijector.onehot_embedder_from_spec(
          spec, dtype=float_dtype, pad_oovs=pad_oovs
      )
    else:
      self.onehot_encoder = ModelInputArrayBijector.identity(spec)

    spec = self.onehot_encoder.output_spec
    self._output_spec = spec
    self._should_clip = should_clip

  def convert(self, trials: Sequence[pyvizier.TrialSuggestion]) -> np.ndarray:
    """Returns an array of shape [len(trials), output_spec.num_dimensions].

    Args:
      trials:

    Returns:
      For each `trial`, if `self.getter(trial)` returns `None`, we _impute_ the
      value; otherwise, we _extract_ the value.
      If `self.parameter_config.type` is `DOUBLE`, then
        * EXTRACT: Directly use the getter's return value as float.
        * IMPUTE: Return `nan`.
      Otherwise,
        * EXTRACT: Returns the integer index of the getter's return value within
          feasible values.
        * IMPUTE: Returns `len(feasible_values)`.
    """
    if not trials:
      return np.zeros(
          [0, self.output_spec.num_dimensions], dtype=self.output_spec.dtype
      )

    value_converter = (
        self._convert_index
        if self._getter_spec.type == NumpyArraySpecType.DISCRETE
        else self._convert_continuous
    )
    values = [value_converter(t) for t in trials]
    array = np.asarray(values, dtype=self._getter_spec.dtype).reshape([-1, 1])
    return self.onehot_encoder.forward_fn(self.scaler.forward_fn(array))

  def _to_parameter_value(
      self, value: Union['np.float', float, int]
  ) -> Optional[pyvizier.ParameterValue]:
    """Converts to a single parameter value; see to_parameter_values().

    Be aware that the value is automatically truncated.

    Args:
      value:

    Returns:
      ParameterValue.
    """
    if not self._converts_to_parameter:
      return None
    elif not np.isfinite(value):
      return None
    elif self.parameter_config.type == pyvizier.ParameterType.DOUBLE:
      # Input parameter was DOUBLE. Output is also DOUBLE.
      if self._should_clip:
        value = np.clip(
            value,
            self._parameter_config.bounds[0],
            self._parameter_config.bounds[1],
        )
      return pyvizier.ParameterValue(float(value))
    elif self.output_spec.type == NumpyArraySpecType.CONTINUOUS:
      # The parameter config is originally discrete, but continuified.
      # Round to the closest number.
      diffs = np.abs(
          np.asarray(self.parameter_config.feasible_values, dtype=self.dtype)
          - value
      )

      idx = np.argmin(diffs)
      closest_number = pyvizier.ParameterValue(
          self.parameter_config.feasible_values[idx]
      )
      return pyvizier.ParameterValue(
          closest_number.cast_as_internal(self.parameter_config.type)
      )

    elif value >= len(self.parameter_config.feasible_values):
      return None
    else:
      return pyvizier.ParameterValue(
          self.parameter_config.feasible_values[value]
      )

  def to_parameter_values(
      self, array: np.ndarray
  ) -> List[Optional[pyvizier.ParameterValue]]:
    """Convert and clip to the nearest feasible parameter values."""
    array = self.scaler.backward_fn(self.onehot_encoder.backward_fn(array))
    return [self._to_parameter_value(v) for v in list(array.flatten())]

  def _convert_index(self, trial: pyvizier.TrialSuggestion):
    """Called by `convert()` if configured for a non-continuous parameter."""
    raw_value = self._getter(trial)
    if raw_value in self.parameter_config.feasible_values:
      return self.parameter_config.feasible_values.index(raw_value)
    else:
      # Return the catch-all missing index.
      return len(self.parameter_config.feasible_values)

  def _convert_continuous(self, trial: pyvizier.TrialSuggestion):
    """Called by `convert()` if configured for a continuous parameter."""
    raw_value = self._getter(trial)
    if raw_value is None:
      return np.nan
    else:
      return raw_value

  @property
  def dtype(self):
    """dtype of the array returned by convert()."""
    return self.output_spec.dtype

  @property
  def output_spec(self) -> NumpyArraySpec:
    return self._output_spec

  @property
  def parameter_config(self) -> pyvizier.ParameterConfig:
    """Original parameter config that this converter is defined on."""
    return self._parameter_config


class ModelOutputConverter(metaclass=abc.ABCMeta):
  """Metric converter interface.

  Unlike ModelInputConverter, this class is currently designed to always
  output a single-dimensional metric. This is because in standard Keras
  workflows, output shapes must be known when defining a model. (In contrast,
  input shapes are required at the time of _building_ a model.)
  """

  @abc.abstractmethod
  def convert(self, measurements: Sequence[pyvizier.Measurement]) -> np.ndarray:
    """Returns N x 1 array."""
    pass

  @abc.abstractmethod
  def to_metrics(
      self, labels: np.ndarray
  ) -> Sequence[Optional[pyvizier.Metric]]:
    """Returns a list of pyvizier metrics.

    The metrics are converted from an array of labels with shape (len(metrics),)
    or (len(metrics), 1) and nan values in the labels are translated to None.

    Args:
      labels: (len(metrics),) or (len(metrics), 1) shaped array of labels.

    Returns:
      A list of pyvizier metrics created with `labels`.
    """

  @property
  @abc.abstractmethod
  def metric_information(self) -> pyvizier.MetricInformation:
    """Describes the semantics of the return value from convert() method.

    Returns:
      The MetricInformation that reflects how the labels should be
      interpreted. It may not be identical to the MetricInformation that the
      converter was created from. For example, if the converter flips the signs
      or changes the semantics of safety configs, then the returned
      MetricInformation should reflect such changes.
    """

  @property
  def output_shape(self) -> Tuple[None, int]:
    return (None, 1)


class DefaultModelOutputConverter(ModelOutputConverter):
  """Converts measurements into numpy arrays."""

  def __init__(
      self,
      metric_information: pyvizier.MetricInformation,
      *,
      flip_sign_for_minimization_metrics: bool = False,
      shift_safe_metrics: bool = True,
      dtype: Union[
          Type[float], Type[int], Type[np.generic], np.dtype
      ] = np.float32,
      raise_errors_for_missing_metrics: bool = False,
  ):
    """Init.

    Args:
      metric_information:
      flip_sign_for_minimization_metrics: Flips the sign if the metric is to
        minimize.
      shift_safe_metrics: If True, subtract the safety threshold from the metric
        value. i.e. center the metric so that the safety threshold is 0. For a
        minimization safety metric if flip_sign_for_minimization_metrics is
        true, subtraction happens begore flipping the sign.
      dtype:
      raise_errors_for_missing_metrics: If True, raise errors. Otherwise, fill
        with NaN.
    """

    self._original_metric_information = metric_information
    self.flip_sign_for_minimization_metrics = flip_sign_for_minimization_metrics
    self.raise_errors_for_missing_metrics = raise_errors_for_missing_metrics
    self.dtype = dtype
    self.shift_safe_metrics = shift_safe_metrics

  @property
  def _should_flip_sign(self) -> bool:
    return (
        self._original_metric_information.goal
        == pyvizier.ObjectiveMetricGoal.MINIMIZE
        and self.flip_sign_for_minimization_metrics
    )

  def convert(
      self, measurements: Sequence[Optional[pyvizier.Measurement]]
  ) -> np.ndarray:
    """Returns a (len(measurements), 1) array."""
    if not measurements:
      return np.zeros([0, 1], dtype=self.dtype)

    all_metrics = [m.metrics if m is not None else dict() for m in measurements]
    if not self.raise_errors_for_missing_metrics:
      metricvalues = [
          metrics.get(self._original_metric_information.name, None)
          for metrics in all_metrics
      ]
      labels = [mv.value if mv else np.nan for mv in metricvalues]
    else:
      labels = [
          metrics[self._original_metric_information.name].value
          for metrics in all_metrics
      ]
    labels = np.asarray(labels, dtype=self.dtype)[:, np.newaxis]
    if (
        self.shift_safe_metrics
        and self._original_metric_information.type.is_safety
    ):
      labels -= self._original_metric_information.safety_threshold
    return labels * (-1 if self._should_flip_sign else 1)

  def to_metrics(
      self, labels: np.ndarray
  ) -> Sequence[Optional[pyvizier.Metric]]:
    """Converts an array of labels to pyvizier Metrics.

    Args:
      labels: (len(metrics),) or (len(metrics), 1) shaped array of labels.

    Returns:
      metrics: a list of pyvizier metrics.
    """
    if labels.ndim == 1:
      labels = labels[:, None]
    if labels.shape[1] > 1 or len(labels.shape) > 2:
      raise ValueError('The input array must be of shape (num,) or (num, 1).')

    labels = labels.flatten()
    if (
        self.shift_safe_metrics
        and self._original_metric_information.type.is_safety
    ):
      labels -= self._original_metric_information.safety_threshold

    labels = labels * (-1 if self._should_flip_sign else 1)
    metrics = [
        pyvizier.Metric(value=l) if np.isfinite(l) else None for l in labels
    ]
    return metrics

  @property
  def metric_information(self) -> pyvizier.MetricInformation:
    """Returns a copy that reflects how the converter treats the metric."""
    metric_information = copy.deepcopy(self._original_metric_information)
    if self.shift_safe_metrics and metric_information.type.is_safety:
      metric_information.safety_threshold = 0.0
    if self._should_flip_sign:
      metric_information = metric_information.flip_goal()
    return metric_information


class DefaultTrialConverter(TrialToNumpyDict):
  """Combines parameter and metric converters.

  Always takes the final measurement and convert to metrics.

  WARNING: This class is not responsible for filtering out non-completed or
  infeasible trials. Labels may contain NaN in such cases.

  For standard use cases, use one of the factory functions. For more
  flexibility, inject parameter_converters and metric_converters on your own.
  """

  def __init__(
      self,
      parameter_converters: Collection[ModelInputConverter],
      metric_converters: Collection[ModelOutputConverter] = tuple(),
  ):
    self.parameter_converters = list(parameter_converters)
    self.parameter_converters_dict = {
        pc.parameter_config.name: pc for pc in self.parameter_converters
    }
    self.metric_converters = list(metric_converters)
    self._metric_converters_dict = {
        mc.metric_information.name: mc for mc in self.metric_converters
    }

  def to_features(
      self, trials: Sequence[pyvizier.TrialSuggestion]
  ) -> Dict[str, np.ndarray]:
    """Shorthand for to_xy(trials))[0]."""
    result_dict = dict()
    for converter in self.parameter_converters:
      result_dict[converter.parameter_config.name] = converter.convert(trials)
    return result_dict

  def to_trials(
      self,
      features: Mapping[str, np.ndarray],
      labels: Optional[Mapping[str, np.ndarray]] = None,
  ) -> List[pyvizier.Trial]:
    """Inverse of `to_features` and optionally the inverse of `to_labels`.

    We assume that label values are already normalized and their signs are
    flipped if required.

    Args:
      features: A dictionary where keys correspond to parameter names in the
        returned Trial and values correspond to parameter values and have shape
        (num_obs, 1).
      labels: A dictionary of labels where each key corresponds to a metric name
        and its value is an array of shape (num_obs, 1) of observed metric
        values.

    Returns:
      A list of pyvizier trials created with parameters corresponding to
      `features` and final measurements corresponding to `labels`.
      NOTE: All final measurements have steps=1. If 'labels' is not passed, the
      output trials include `final_measurement=None` and `measurements=[]`.
    """
    invalid_features = [
        (k, v.shape)
        for k, v in features.items()
        if len(v.shape) != 2 or (len(v.shape) == 2 and v.shape[1] != 1)
    ]
    if invalid_features:
      raise ValueError(
          'Features need to contain 2d arrays with shape (num_obs, 1). Invalid'
          f' feature shapes: {invalid_features}'
      )

    if labels:
      invalid_labels = [
          (k, v.shape)
          for k, v in labels.items()
          if len(v.shape) != 2 or (len(v.shape) == 2 and v.shape[1] != 1)
      ]
      if invalid_labels:
        raise ValueError(
            'Labels need to contain 2d arrays with shape (num_obs, 1). Invalid'
            f' label shapes: {invalid_labels}'
        )

    if labels is None:
      return [
          pyvizier.Trial(parameters=p) for p in self.to_parameters(features)
      ]

    try:
      labels = DictOf2DArrays(labels)
    except ValueError as e:
      raise ValueError('Please check the shape of "labels"') from e

    try:
      features = DictOf2DArrays(features)
    except ValueError as e:
      raise ValueError('Please check the shape of "features"') from e

    if labels.size != features.size:
      raise ValueError(
          'The number of features and labels observations do not match.'
      )
    parameters = self.to_parameters(features)
    measurements = self._to_measurements(labels)
    for m in measurements:
      m.steps = 1

    trials = []
    for p, m in zip(parameters, measurements):
      trial = pyvizier.Trial(parameters=p, final_measurement=m)
      # _to_measurements returns an empty dict for NaN and non-finite metric
      # values.
      if not m.metrics:
        trial.complete(m, infeasibility_reason='Metrics are empty')
      trials.append(trial)
    return trials

  def _to_measurements(
      self, labels: Mapping[str, np.ndarray]
  ) -> List[pyvizier.Measurement]:
    """Converts a dictionary of labels into a list of pyvizier measurements.

    Each key in the dictionary corresponds to a metric and the length of the
    list of pyvizier measurements equals to the number of keys. Note that this
    method generates measurements without setting steps, hence steps are
    defaulted to zero. The caller should assign them later if desired.

    Args:
      labels: A dictionary of labels where each key corresponds to a metric name
        with dictionary values as metric values casted as an array of shape
        (num_obs,) or (num_obs, 1).

    Returns:
      A list of pyvizier measurements created with final measurements
        corresponding to `labels`.
    """
    try:
      DictOf2DArrays(labels)
    except ValueError as e:
      raise ValueError('Please check the shape of "labels"') from e
    measurements = [
        pyvizier.Measurement() for _ in range(DictOf2DArrays(labels).size)
    ]
    # Iterate through labels names and convert them.
    for metric_name, metric_values in labels.items():
      metric_converter = self._metric_converters_dict[metric_name]
      metrics_values = metric_converter.to_metrics(metric_values)
      for measurement_dict, value in zip(measurements, metrics_values):
        if value is not None:
          measurement_dict.metrics[metric_name] = value
    return measurements

  def to_parameters(
      self, features: Mapping[str, np.ndarray]
  ) -> List[pyvizier.ParameterDict]:
    """Convert to nearest feasible parameter value. NaNs trigger errors."""
    # TODO: NaNs should be ignored instead of triggering errors.

    # Validate features's shape, and create empty ParameterDicts.
    parameters = [
        pyvizier.ParameterDict() for _ in range(DictOf2DArrays(features).size)
    ]

    # Iterate through parameter names and convert them.
    for key, values in features.items():
      parameter_converter = self.parameter_converters_dict[key]
      parameter_values = parameter_converter.to_parameter_values(values)
      for param_dict, value in zip(parameters, parameter_values):
        if value is not None:
          param_dict[key] = value
    return parameters

  def to_labels(
      self, trials: Sequence[pyvizier.Trial]
  ) -> Dict[str, np.ndarray]:
    """Shorthand for to_xy(trials))[1]."""
    result_dict = dict()
    for converter in self.metric_converters:
      result_dict[converter.metric_information.name] = converter.convert(
          [t.final_measurement for t in trials]
      )
    return result_dict

  def to_labels_array(self, trials: Sequence[pyvizier.Trial]) -> np.ndarray:
    """Shorthand for dict_to_array(self.to_labels(trials))."""
    return dict_to_array(self.to_labels(trials))

  def to_xy(
      self, trials: Sequence[pyvizier.Trial]
  ) -> Tuple[Dict[str, np.ndarray], Dict[str, np.ndarray]]:
    """See base class."""
    return self.to_features(trials), self.to_labels(trials)

  @property
  def features_shape(self) -> Dict[str, Tuple[Union[int, None], int]]:
    """See base class."""
    return {
        pc.output_spec.name: (None, pc.output_spec.num_dimensions)
        for pc in self.parameter_converters
    }

  @property
  def output_specs(self) -> Dict[str, NumpyArraySpec]:
    """See base class."""
    return {
        pc.output_spec.name: pc.output_spec for pc in self.parameter_converters
    }

  @property
  def labels_shape(self) -> Dict[str, Sequence[Union[int, None]]]:
    """See base class."""
    return {
        mc.metric_information.name: mc.output_shape
        for mc in self.metric_converters
    }

  @property
  def metric_information(self) -> Dict[str, pyvizier.MetricInformation]:
    """See base class."""
    return {
        mc.metric_information.name: mc.metric_information
        for mc in self.metric_converters
    }

  # TODO: Deprecate or update so that it returns SearchSpace.
  @property
  def parameter_configs(self) -> Dict[str, pyvizier.ParameterConfig]:
    """Returns a dict of the original Parameter configs."""
    return {
        converter.parameter_config.name: converter.parameter_config
        for converter in self.parameter_converters
    }

  @classmethod
  def from_study_configs(
      cls,
      study_configs: Sequence[pyvizier.ProblemStatement],
      metric_information: Collection[pyvizier.MetricInformation],
      *,
      use_study_id_feature: bool = True,
  ) -> 'DefaultTrialConverter':
    """Creates a converter from a list of study configs.

    Args:
      study_configs: StudyConfigs to be merged.
      metric_information: MetricInformation of metrics to be used as y-values.
      use_study_id_feature: If True, an extra parameter is added that
        corresponds to the STUDY_ID_FIELD inside metadata.

    Returns:
      `DefaultTrialConverter`.
    """
    # Merge parameter configs by name.
    merged_configs = list(
        pyvizier.SearchSpaceSelector([sc.search_space for sc in study_configs])
        .select_all()
        .merge()
    )

    merged_configs = {pc.name: pc for pc in merged_configs}
    parameter_converters = [
        DefaultModelInputConverter(pc) for pc in merged_configs.values()
    ]

    # Append study id feature if configured to do so.
    if use_study_id_feature:
      # Collect study ids.
      study_ids = set()
      for sc in study_configs:
        metalearn_study_id = sc.metadata.get(STUDY_ID_FIELD, None)
        if metalearn_study_id is None:
          continue
        study_ids.add(metalearn_study_id)

      # Validate.
      if not study_ids:
        logging.error(
            'use_study_id_feature was True, but none of the studies '
            'had study id configured.'
        )
        use_study_id_feature = False
      elif STUDY_ID_FIELD in merged_configs:
        raise ValueError(
            'Dataset name conflicts with a ParameterConfig '
            'that already exists: {}'.format(merged_configs[STUDY_ID_FIELD])
        )

      # Create new parameter config.
      parameter_config = pyvizier.ParameterConfig.factory(
          STUDY_ID_FIELD, feasible_values=list(study_ids)
      )
      merged_configs[STUDY_ID_FIELD] = parameter_config
      logging.info('Created a new ParameterConfig %s', parameter_config)

      # Create converter.
      parameter_converters.append(
          DefaultModelInputConverter(
              parameter_config,
              lambda t: t.metadata.get(STUDY_ID_FIELD, None),
              converts_to_parameter=False,
          )
      )

    return cls(
        parameter_converters,
        [DefaultModelOutputConverter(m) for m in metric_information],
    )

  @classmethod
  def from_study_config(
      cls, study_config: pyvizier.ProblemStatement
  ) -> 'DefaultTrialConverter':
    return cls.from_study_configs(
        [study_config],
        study_config.metric_information,
        use_study_id_feature=False,
    )


@attr.define
class TrialToArrayConverter:
  """EXPERIMENTAL: A quick-and-easy converter that returns a single array.

  Unlike TrialtoNumpyDict converters, `to_features` and `to_labels`
  return a single array of floating numbers. CATEGORICAL and DISCRETE parameters
  are one-hot embedded.

  IMPORTANT: Use a factory method (currently, there is one: `from_study_config`)
  instead of `__init__`.

  WARNING: This class is not exchangable with `TrialToNumpyDict`, and does not
  have functions that return shape or metric informations. Use it at your own
  risk.
  """

  _impl: DefaultTrialConverter

  def to_features(self, trials) -> np.ndarray:
    """Returns the features array with dimension: (n_trials, n_features)."""
    return dict_to_array(self._impl.to_features(trials))

  def to_labels(self, trials) -> np.ndarray:
    """Returns the labels array with dimension: (n_trials, n_metrics)."""
    # Pad up the labels.
    return dict_to_array(self._impl.to_labels(trials))

  def to_xy(self, trials) -> Tuple[np.ndarray, np.ndarray]:
    return self.to_features(trials), self.to_labels(trials)

  def to_parameters(self, arr: np.ndarray) -> Sequence[pyvizier.ParameterDict]:
    """Convert to nearest feasible parameter value. NaNs are preserved."""
    arrformat = DictOf2DArrays(self._impl.to_features([]))
    return self._impl.to_parameters(arrformat.dict_like(arr))

  @classmethod
  def from_study_config(
      cls,
      study_config: pyvizier.ProblemStatement,
      *,
      scale: bool = True,
      pad_oovs: bool = True,
      max_discrete_indices: int = 0,
      flip_sign_for_minimization_metrics: bool = True,
      should_clip=True,
      dtype=np.float64,
  ) -> 'TrialToArrayConverter':
    """From study config.

    Args:
      study_config:
      scale: If True, scales the parameters to [0, 1] range.
      pad_oovs: If True, add an extra dimension for out-of-vocabulary values for
        non-CONTINUOUS parameters.
      max_discrete_indices: For DISCRETE and INTEGER parameters that have more
        than this many feasible values will be continuified. When generating
        suggestions, values are rounded to the nearest feasible value. Note this
        default is different from the default in DefaultModelInputConverter.
      flip_sign_for_minimization_metrics: If True, flips the metric signs so
        that every metric maximizes.
      should_clip: Whether or not clipping should be done.
      dtype: dtype

    Returns:
      TrialToArrayConverter
    """

    def create_input_converter(parameter):
      return DefaultModelInputConverter(
          parameter,
          scale=scale,
          max_discrete_indices=max_discrete_indices,
          onehot_embed=True,
          float_dtype=dtype,
          pad_oovs=pad_oovs,
          should_clip=should_clip,
      )

    def create_output_converter(metric):
      return DefaultModelOutputConverter(
          metric,
          flip_sign_for_minimization_metrics=flip_sign_for_minimization_metrics,
          dtype=dtype,
      )

    sc = study_config  # alias, to keep pylint quiet in the next line.
    converter = DefaultTrialConverter(
        [
            create_input_converter(p)
            for p in sc.search_space.root.select_all().merge()
        ],
        [create_output_converter(m) for m in sc.metric_information],
    )
    return cls(converter)

  @property
  def output_specs(self) -> Sequence[NumpyArraySpec]:
    return [
        converter.output_spec for converter in self._impl.parameter_converters
    ]

  @property
  def metric_specs(self) -> Sequence[pyvizier.MetricInformation]:
    return [mc.metric_information for mc in self._impl.metric_converters]

  @property
  def dtype(self) -> np.dtype:
    return self.to_features([]).dtype



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Eagle meta learner search space.
"""

from vizier import pyvizier as vz


def meta_eagle_search_space() -> vz.SearchSpace:
  """Returns the meta eagle search space."""
  search_space = vz.SearchSpace()
  # Perturbation
  search_space.root.add_float_param(
      name='perturbation',
      min_value=1e-4,
      max_value=1e2,
      default_value=1e-1,
      scale_type=vz.ScaleType.LOG,
  )
  search_space.root.add_float_param(
      name='perturbation_lower_bound',
      min_value=1e-5,
      max_value=1e-1,
      default_value=1e-3,
      scale_type=vz.ScaleType.LOG,
  )
  # Gravity
  search_space.root.add_float_param(
      name='gravity',
      min_value=1e-2,
      max_value=1e2,
      default_value=1.0,
      scale_type=vz.ScaleType.LOG,
  )
  # Visibility
  search_space.root.add_float_param(
      name='visibility',
      min_value=3 * 1e-2,
      max_value=3 * 1e2,
      default_value=3.0,
      scale_type=vz.ScaleType.LOG,
  )
  search_space.root.add_float_param(
      name='categorical_visibility',
      min_value=2.0 * 1e-3,
      max_value=2.0 * 1e1,
      default_value=2.0 * 1e-1,
      scale_type=vz.ScaleType.LOG,
  )
  search_space.root.add_float_param(
      name='discrete_visibility',
      min_value=1e-2,
      max_value=1e2,
      default_value=1.0,
      scale_type=vz.ScaleType.LOG,
  )
  search_space.root.add_float_param(
      name='categorical_perturbation_factor',
      min_value=2.5 * 1e-1,
      max_value=2.5 * 1e3,
      default_value=2.5 * 1e1,
      scale_type=vz.ScaleType.LOG,
  )
  search_space.root.add_float_param(
      name='discrete_perturbation_factor',
      min_value=1e-1,
      max_value=1e3,
      default_value=1e1,
      scale_type=vz.ScaleType.LOG,
  )
  # Pool size.
  search_space.root.add_float_param(
      name='pool_size_factor',
      min_value=1.0,
      max_value=2.0,
      default_value=1.2,
      scale_type=vz.ScaleType.LOG,
  )
  search_space.root.add_float_param(
      name='negative_gravity',
      min_value=2.0 * 1e-4,
      max_value=2.0,
      default_value=2 * 1e-2,
      scale_type=vz.ScaleType.LOG,
  )
  search_space.root.add_float_param(
      name='pure_categorical_perturbation',
      min_value=1e-3,
      max_value=1e1,
      default_value=1e-1,
      scale_type=vz.ScaleType.LOG,
  )

  return search_space



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Scalarizations that weights multiple metrics into a scalar."""

import abc
from typing import Callable, Optional

import equinox as eqx
import jax
from jax import numpy as jnp
import jaxtyping as jt
import typeguard


class Scalarization(abc.ABC, eqx.Module):
  """Reduces an array of objectives to a single float.

  Assumes all objectives are for MAXIMIZATION.
  """

  # Weights shape should be broadcastable with objectives when called.
  weights: jt.Float[jax.Array, '#Obj'] = eqx.field(converter=jnp.asarray)

  @jt.jaxtyped(typechecker=typeguard.typechecked)
  def __call__(
      self, objectives: jt.Float[jax.Array, '*Batch Obj']
  ) -> jt.Float[jax.Array, '*Batch']:
    pass


# Scalarization factory from weights.
ScalarizationFromWeights = Callable[
    [jt.Float[jax.Array, '#Obj']], Scalarization
]


class LinearScalarization(Scalarization):
  """Linear Scalarization."""

  @jt.jaxtyped(typechecker=typeguard.typechecked)
  def __call__(
      self, objectives: jt.Float[jax.Array, '*Batch Obj']
  ) -> jt.Float[jax.Array, '*Batch']:
    return jnp.sum(self.weights * objectives, axis=-1)


class ChebyshevScalarization(Scalarization):
  """Chebyshev Scalarization."""

  @jt.jaxtyped(typechecker=typeguard.typechecked)
  def __call__(
      self, objectives: jt.Float[jax.Array, '*Batch Obj']
  ) -> jt.Float[jax.Array, '*Batch']:
    return jnp.min(objectives * self.weights, axis=-1)


class HyperVolumeScalarization(Scalarization):
  """HyperVolume Scalarization."""

  reference_point: Optional[jt.Float[jax.Array, '* #Obj']] = eqx.field(
      default=None
  )

  @jt.jaxtyped(typechecker=typeguard.typechecked)
  def __call__(
      self, objectives: jt.Float[jax.Array, '*Batch Obj']
  ) -> jt.Float[jax.Array, '*Batch']:
    # Uses scalarizations in https://arxiv.org/abs/2006.04655 for
    # non-convex multiobjective optimization. Removes the exponentiation
    # factor in number of objectives as it is a monotone transformation and
    # removes the non-negativity for easier gradients.
    if self.reference_point is not None:
      return jnp.min(
          (objectives - self.reference_point) / self.weights, axis=-1
      )
    else:
      return jnp.min(objectives / self.weights, axis=-1)


class LinearAugmentedScalarization(Scalarization):
  """Scalarization augmented with a linear sum.

  See https://arxiv.org/pdf/1904.05760.pdf.
  """

  scalarization_factory: ScalarizationFromWeights = eqx.field(static=True)
  augment_weight: jt.Float[jax.Array, ''] = eqx.field(
      default=1.0, converter=jnp.asarray
  )

  @jt.jaxtyped(typechecker=typeguard.typechecked)
  def __call__(
      self, objectives: jt.Float[jax.Array, '*Batch Obj']
  ) -> jt.Float[jax.Array, '*Batch']:
    return self.scalarization_factory(self.weights)(
        objectives
    ) + self.augment_weight * LinearScalarization(weights=self.weights)(
        objectives
    )



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Cross-Platform Backend."""
import enum
import functools
import getpass
import random
import threading
import time
from typing import Any, Dict, Optional, Sequence, Type, Union

from absl import logging
import attrs
import pyglove as pg
from vizier import pyvizier as vz
from vizier._src.pyglove import algorithms
from vizier._src.pyglove import client
from vizier._src.pyglove import constants
from vizier._src.pyglove import converters
from vizier._src.pyglove import core
from vizier._src.pyglove import pythia as pyglove_pythia
from vizier.client import client_abc


TunerPolicy = pyglove_pythia.TunerPolicy
BuiltinAlgorithm = algorithms.BuiltinAlgorithm

ExpandedStudyName = client.ExpandedStudyName
StudyKey = client.StudyKey


class TunerMode(enum.Enum):
  """Mode for Tuner."""

  # Automatic select primary mode when study is new or study TUNER_ID equals
  # to current tuner ID (for failover) or the primary tuner is not accessible
  # via its BNS. Otherwise select secondary mode.
  AUTO = 0

  # Work as primary tuner, which will host pythia service when algorithm is
  # non-Vizier-builtin. When using Vizier built-in algorithms, all tuners are
  # in secondary mode.
  PRIMARY = 1

  # Work as secondary tuner, which can query or stop a tuner task but not
  # hosting pythia service.
  SECONDARY = 2


# Global policy cache.
_global_policy_cache: Dict[StudyKey, TunerPolicy] = dict()


@attrs.define(auto_attribs=False)
class VizierBackend(pg.tuning.Backend):
  """Vizier backend."""

  # Class-level variables.
  default_owner: str = getpass.getuser()
  default_study_prefix: Optional[str] = None
  tuner_cls: Type[client.VizierTuner] = attrs.field()

  # Instance-level variables.

  #
  # Flags passed from `pg.sample`:
  #

  # Worker group - workers that belong to the same group will share the same
  # Vizier client ID.
  _group: Union[None, int, str] = attrs.field()

  # Max number of examples to sample.
  _num_examples: Optional[int] = attrs.field()

  # Prior study IDs for transfer learning.
  _prior_study_ids: Optional[Sequence[str]] = attrs.field()

  # If True, add the completed trials from prior studies. Otherwise, simply
  # warm up the algorithm using these trials without adding them to the current
  # study.
  _add_prior_trials: bool = attrs.field()

  #
  # Internal states.
  #

  _tuner: client.VizierTuner
  _dna_spec: pg.DNASpec
  _algorithm: pg.geno.DNAGenerator = attrs.field()
  _early_stopping_policy: Optional[pg.tuning.EarlyStoppingPolicy] = (
      attrs.field()
  )

  _study_owner: str = attrs.field()
  _study_name: ExpandedStudyName = attrs.field()
  _converter: converters.VizierConverter = attrs.field()

  _study: client_abc.StudyInterface = attrs.field()
  _suggestion_generator: Any = attrs.field()

  _run_mode: TunerMode = attrs.field()
  _auto_election_thread: Optional[threading.Thread] = attrs.field()
  _is_active: bool = attrs.field()

  def __init__(
      self,
      name: Optional[str],
      group: Union[None, int, str],
      dna_spec: pg.DNASpec,
      algorithm: pg.DNAGenerator,
      metrics_to_optimize: Sequence[str],
      early_stopping_policy: Optional[pg.tuning.EarlyStoppingPolicy] = None,
      num_examples: Optional[int] = None,
      study_owner: Optional[str] = None,
      prior_study_ids: Optional[Sequence[str]] = None,
      add_prior_trials: bool = False,
      is_chief: Optional[bool] = None,
  ):
    self._tuner = self.tuner_cls()

    self._dna_spec = dna_spec
    self._algorithm = algorithm
    self._early_stopping_policy = early_stopping_policy

    self._group = group
    self._num_examples = num_examples
    self._prior_study_ids = prior_study_ids or tuple()
    self._add_prior_trials = add_prior_trials

    self._study_owner = study_owner or self.default_owner
    self._study_name = self._expand_name(name)

    # Set up converter based on the search space and metrics to optimize.
    self._converter = converters.VizierConverter.from_dna_spec(
        dna_spec, metrics_to_optimize
    )

    # Detecting mode based on is_chef flag.
    if is_chief:
      mode = TunerMode.PRIMARY
    elif is_chief is None:
      mode = TunerMode.AUTO
    else:
      mode = TunerMode.SECONDARY

    self._run_mode = mode

    # Setup the Vizier study and the suggestion generator.
    is_chief = self._setup_study()
    self._suggestion_generator = self._create_suggestion_generator()

    # Start Pythia if needed.
    self._auto_election_thread = None
    self._is_active = True

    if self._need_pythia_service:
      pg.logging.info(
          "Study '%s/%s' will be served by Pythia service hosted on %r. "
          'Algorithm=%r, EarlyStoppingPolicy=%r.',
          self._study_owner,
          self._study_name,
          self._get_chief_tuner_id(),
          self._algorithm,
          self._early_stopping_policy,
      )

      if is_chief:
        pg.logging.info(
            "Starting hosting Pythia service for study '%s/%s' as PRIMARY. ",
            self._study_owner,
            self._study_name,
        )
        self._start_pythia()
      if self._run_mode == TunerMode.AUTO:
        self._auto_election_thread = threading.Thread(
            target=self._auto_elect_primary_if_needed
        )
        self._auto_election_thread.start()
    else:
      pg.logging.info(
          'Pythia service is not required as both the search algorithm and '
          'early stopping policy are either Vizier built-in or served from '
          'remote Pythia endpoint. Algorithm=%r, EarlyStoppingPolicy=%r.',
          self._algorithm,
          self._early_stopping_policy,
      )

  @property
  def _host_pythia_algorithm(self) -> bool:
    """Returns True if the algorithm is hosted on Pythia."""
    return not isinstance(self._algorithm, algorithms.PseudoAlgorithm)

  @property
  def _need_pythia_service(self) -> bool:
    """Returns True if pythia service is needed."""
    return self._host_pythia_algorithm

  def _setup_study(self) -> bool:
    """Sets up Vizier study, returns True if current worker is the chief."""
    study_descriptor = f'{self._study_owner}/{self._study_name}'

    try:
      self._study = self._tuner.load_study(self._study_owner, self._study_name)
      pg.logging.info(
          'Connecting tuner to existing study %r... ', study_descriptor
      )

      #
      # Ensure the client-side search space matches with the server-side search
      # space.
      #
      stored_dna_spec = self._get_stored_dna_spec()

      # CustomDecisionPoint could be non-serializable and non-compariable.
      if pg.contains(self._dna_spec, pg.geno.CustomDecisionPoint):
        pg.logging.info(
            'There is a CustomDecisionPoint. Skipping the check to ensure the '
            'client-side search space matches the server-side search space.'
        )
      elif pg.eq(self._dna_spec, stored_dna_spec):
        pg.logging.info(
            'Verified that the client-side search space matches the '
            'server-side.'
        )
      else:
        raise ValueError(
            'The client-side search space is different from the search space '
            'from the study stored in Vizier. Try launching the experiment '
            'with a different study name. '
            f'Diff: {pg.diff(stored_dna_spec, self._dna_spec)}.'
        )

      chief_tuner_id = self._get_chief_tuner_id()

      if self._run_mode != TunerMode.SECONDARY and not self._tuner.ping_tuner(
          chief_tuner_id
      ):
        # NOTE(daiyip): there could be a race condition that multiple workers
        # elect themselves as the new primary, and all of them regard
        # themselves as the elected. When this happens, multiple workers
        # may be hosting the Pythia service for this study. However, the study
        # will use one address (BNS of the latest elected worker) as the
        # Pythia endpoint. Besides, all states are stored in the study,
        # restart of workers will pick up these states from the study.
        # This is done by replaying existing trials to recompute the states of
        # the search algorithm. Therefore, it does not matter which worker is
        # chosen to serve the study, which should work equally well.
        # On the other hand, it is cheap to serve a PythiaService without
        # incoming queries. Therefore, we do not handle this race conditions
        # with expensive distributed locks.
        chief_tuner_id = self._register_self_as_primary()

      is_chief = chief_tuner_id == self._tuner_id
      if self._run_mode == TunerMode.SECONDARY and is_chief:
        raise ValueError(
            f'Inconsistent primary tuner: {self._tuner_id!r} is running as '
            f'secondary but study {study_descriptor!r} indicates otherwise.'
        )
      elif self._run_mode == TunerMode.PRIMARY and not is_chief:
        raise ValueError(
            f'Inconsistent tuner mode: {self._tuner_id!r} is running in '
            f'PRIMARY mode but study {study_descriptor!r} already has a '
            f'different primary {chief_tuner_id!r}.\n'
            'Please check if you are launching different experiments using '
            'the same Vizier study.'
        )

      pg.logging.info(
          'Tuner is running in %s mode with existing study %r.',
          'PRIMARY' if is_chief else 'SECONDARY',
          study_descriptor,
      )

      return is_chief
    except KeyError:
      # Study does not exist.
      if self._run_mode == TunerMode.SECONDARY:
        pg.logging.info(
            'Start tuner as secondary. Waiting for study %r to be created...',
            study_descriptor,
        )
        self._study = self._wait_for_study(self._study_owner, self._study_name)
      else:
        mode_str = 'PRIMARY' if self._run_mode == TunerMode.PRIMARY else 'AUTO'
        pg.logging.info(
            'Start tuner in %s mode. Attempting to create study %r...',
            mode_str,
            study_descriptor,
        )

        # Create a new study.
        problem = self._converter.problem_or_dummy
        problem.metadata.ns(constants.METADATA_NAMESPACE)[
            constants.STUDY_METADATA_KEY_TUNER_ID
        ] = self._tuner_id
        self._study = self._tuner.create_study(
            problem,
            self._converter,
            self._study_owner,
            self._study_name,
            self._algorithm,
            self._early_stopping_policy,
        )

      # Perform post-study actions.
      is_chief = self._tuner_id == self._get_chief_tuner_id()
      if is_chief:
        self._on_study_created()

      pg.logging.info(
          'Tuner is running in %s mode with newly created study %r.',
          'PRIMARY' if is_chief else 'SECONDARY',
          study_descriptor,
      )
      return is_chief

  def _on_study_created(self):
    # For multi-thread scenario, `local_tuner_id` will be the same for
    # all the worker threads, therefore there is a chance that multiple
    # worker threads consider themselves as PRIMARY. This does not matter
    # since there is only one Pythia service shared across them.
    if self._add_prior_trials:
      # Trials are added to the study directly upon creation.
      trials: Sequence[vz.Trial] = self._load_prior_trials()
      for trial in trials:
        self._study._add_trial(trial)  # pylint: disable=protected-access

  def _create_suggestion_generator(self):
    """Creates a suggestion generator."""
    while (
        self._num_examples is None
        or len(list(self._study.trials())) < self._num_examples
    ):
      trials = self._study.suggest(
          count=1, client_id=self._tuner.get_group_id(self._group)
      )
      if not trials:
        return
      for trial in trials:
        yield core.Feedback(trial, self._converter)

  def _start_pythia(self) -> None:
    """Starts Pythia service."""
    # Start pythia service.
    self._tuner.start_pythia_service(_global_policy_cache)

    # Set up the policy.
    self._algorithm.setup(self._dna_spec)
    prior_trials = tuple()
    if not self._add_prior_trials:
      # Trials are added to the algorithm only.
      prior_trials: Sequence[vz.Trial] = self._load_prior_trials()
    policy = self._create_policy(prior_trials)
    # Connect the service with policy.
    self._register(self._study_owner, self._study_name, policy)

  def _create_policy(
      self,
      prior_trials: Sequence[vz.Trial],
  ) -> TunerPolicy:
    """Creates a pythia policy."""
    if prior_trials:

      def get_trial_history(vizier_trials):
        for trial in vizier_trials:
          tuner_trial = core.VizierTrial(self._converter, trial)
          reward = tuner_trial.get_reward_for_feedback(
              self._converter.metrics_to_optimize
          )
          yield (tuner_trial.dna, reward)

      self._algorithm.recover(get_trial_history(prior_trials))

    return TunerPolicy(
        self._tuner.pythia_supporter(self._study),
        self._converter,
        self._algorithm,
        early_stopping_policy=self._early_stopping_policy,
    )

  def _get_stored_dna_spec(self) -> pg.DNASpec:
    metadata = self._study.materialize_problem_statement().metadata.ns(
        constants.METADATA_NAMESPACE
    )
    try:
      return converters.restore_dna_spec(
          metadata[constants.STUDY_METADATA_KEY_DNA_SPEC]
      )
    except KeyError as e:
      raise RuntimeError(
          f'Metadata {constants.STUDY_METADATA_KEY_DNA_SPEC} does not exist '
          f'in study: {self._study.resource_name}.'
      ) from e

  def _get_chief_tuner_id(self) -> str:
    metadata = self._study.materialize_problem_statement().metadata.ns(
        constants.METADATA_NAMESPACE
    )
    try:
      return str(metadata[constants.STUDY_METADATA_KEY_TUNER_ID])
    except KeyError as e:
      raise RuntimeError(
          f'Metadata {constants.STUDY_METADATA_KEY_TUNER_ID} does not exist '
          f'in study: {self._study.resource_name}.'
      ) from e

  @functools.cached_property
  def _tuner_id(self) -> str:
    """Returns the tuner id of current ."""
    return self._tuner.get_tuner_id(self._algorithm)

  def _register_self_as_primary(self) -> str:
    metadata = vz.Metadata()
    metadata.ns(constants.METADATA_NAMESPACE)[
        constants.STUDY_METADATA_KEY_TUNER_ID
    ] = self._tuner_id
    self._study.update_metadata(metadata)
    self._tuner.use_pythia_for_study(self._study)
    return self._tuner_id

  def _auto_elect_primary_if_needed(self) -> None:
    """Automatically elect primary tuner if previous primary is offline."""
    assert self._need_pythia_service
    assert self._run_mode == TunerMode.AUTO

    while self._is_active:
      chief_tuner_id = self._get_chief_tuner_id()
      if not self._tuner.ping_tuner(chief_tuner_id):
        # NOTE(daiyip): there could be a race condition that multiple workers
        # elect themselves as the new primary, and all of them regard
        # themselves as the elected. When this happens, multiple workers
        # may be hosting the Pythia service for this study. However, the study
        # will use one address (BNS of the latest elected worker) as the
        # Pythia endpoint. Besides, all states are stored in the study,
        # restart of workers will pick up these states from the study.
        # This is done by replaying existing trials to recompute the states of
        # the search algorithm. Therefore, it does not matter which worker is
        # chosen to serve the study, which should work equally well.
        # On the other hand, it is cheap to serve a PythiaService without
        # incoming queries. Therefore, we do not handle this race conditions
        # with expensive distributed locks.
        new_chief_tuner_id = self._register_self_as_primary()
        self._is_primary = new_chief_tuner_id == self._tuner_id
        if self._is_primary:
          pg.logging.warning(
              'Primary tuner has been switched from %s to %s.',
              chief_tuner_id,
              new_chief_tuner_id,
          )
          self._start_pythia()
      time.sleep(random.randint(50, 70))

  def next(self) -> pg.tuning.Feedback:
    """Gets the next tuning feedback object."""
    try:
      trial = next(self._suggestion_generator)  # pytype: disable=wrong-arg-types
      return core.Feedback(self._study.get_trial(trial.id), self._converter)
    except StopIteration as e:
      self._is_active = False
      raise e

  def _wait_for_study(
      self, owner: str, name: ExpandedStudyName
  ) -> client_abc.StudyInterface:
    """Wait for the study in a loop."""
    while True:
      try:
        return self._tuner.load_study(owner, name)
      except KeyError:
        logging.info(
            'Study %s (owner=%s) does not exist. Retrying after 10 seconds.',
            name,
            owner,
        )
        time.sleep(10)
      except Exception as e:  # pylint:disable=broad-except
        logging.warn(
            'Could not look up study: %s. Retrying after 60 seconds', e
        )
        time.sleep(60)

  def _load_prior_trials(self) -> list[vz.Trial]:
    trials = []
    for prior in self._prior_study_ids:
      trials.extend(
          self._tuner.load_prior_study(prior)
          .trials(vz.TrialFilter(status=vz.TrialStatus.COMPLETED))
          .get()
      )
    return trials

  def _register(
      self, owner: str, name: ExpandedStudyName, policy: TunerPolicy
  ) -> None:
    """Registers the algorithm for a specific study."""
    study_key = StudyKey(owner, name)

    if study_key in _global_policy_cache:
      existing = _global_policy_cache[study_key]
      if existing.algorithm != policy.algorithm:
        raise ValueError(
            f'Different algorithms are used for the same study {study_key!r}. '
            f'Previous: {existing.algorithm!r}, Current: {policy.algorithm!r}.'
        )
      if existing.early_stopping_policy != policy.early_stopping_policy:
        raise ValueError(
            'Different early stopping policy are used for the same study '
            f'{study_key!r}. Previous: {existing.early_stopping_policy!r}, '
            f'Current: {policy.early_stopping_policy!r}.'
        )

    _global_policy_cache[study_key] = policy

  #
  # Class methods.
  #

  @classmethod
  def use_study_prefix(cls, study_prefix: Optional[str]):
    cls.default_study_prefix = study_prefix or ''

  @classmethod
  def _expand_name(cls, name: Optional[str]) -> ExpandedStudyName:
    """Expand the pyglove study name into the full name in Vizier DB.

    Args:
      name: Name as passed into pyglove.

    Returns:
      Study name to use for Vizier interactions.
    """
    components = []
    if cls.default_study_prefix:
      components.append(cls.default_study_prefix)
    if name:
      components.append(name)
    return ExpandedStudyName('.'.join(components))

  @classmethod
  def _get_study_resource_name(cls, name: str) -> str:
    """Use for testing only."""
    return cls.tuner_cls.load_study(
        owner=cls.default_owner,
        name=ExpandedStudyName(name),
    ).resource_name

  @classmethod
  def poll_result(
      cls, name: str, study_owner: Optional[str] = None
  ) -> pg.tuning.Result:
    """Polls result of a study."""
    return core.Result.from_study(
        cls.tuner_cls.load_study(
            study_owner or cls.default_owner, cls._expand_name(name)
        )
    )



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Analyzers for BenchmarkStates for fast comparisons and statistics."""

import json
from typing import Dict, Optional, Sequence, Tuple
from absl import logging
import attrs
import numpy as np
import pandas as pd
from vizier import benchmarks
from vizier import pyvizier as vz
from vizier._src.benchmarks.analyzers import convergence_curve
from vizier.benchmarks import experimenters

RECORD_OBJECTIVE_KEY = 'objective'


@attrs.define(init=True, kw_only=True)
class PlotElement:
  """PlotElement with relevant information for a subplot."""

  curve: Optional[convergence_curve.ConvergenceCurve] = attrs.field(
      default=None,
      validator=attrs.validators.optional(
          attrs.validators.instance_of(convergence_curve.ConvergenceCurve)
      ),
  )
  plot_array: Optional[np.ndarray] = attrs.field(
      default=None,
      validator=attrs.validators.optional(
          attrs.validators.instance_of(np.ndarray)
      ),
  )
  # Error-bar uses curve, whereas histogram/scatter uses plot_array.
  plot_type: str = attrs.field(
      default='error-bar',
      validator=attrs.validators.in_(['error-bar', 'histogram', 'scatter']),
  )
  yscale: str = attrs.field(
      default='linear',
      validator=attrs.validators.in_(['linear', 'symlog', 'logit']),
  )
  # Lower and upper percentiles to display for error bar.
  percentile_error_bar: Tuple[int, int] = attrs.field(
      default=(25, 75),
      validator=attrs.validators.deep_iterable(
          member_validator=attrs.validators.instance_of(int),
          iterable_validator=attrs.validators.instance_of(tuple),
      ),
  )


# Stores all relevant information and plots for a specific BenchmarkState.
@attrs.define(init=True, kw_only=True)
class BenchmarkRecord:
  algorithm: str = attrs.field(
      default='',
      validator=attrs.validators.instance_of(str),
  )
  experimenter_metadata: vz.Metadata = attrs.field(
      factory=vz.Metadata, validator=attrs.validators.instance_of(vz.Metadata)
  )
  plot_elements: Dict[str, PlotElement] = attrs.field(factory=dict)


class BenchmarkStateAnalyzer:
  """Analyzer for BenchmarkStates."""

  @classmethod
  def to_curve(
      cls,
      states: list[benchmarks.BenchmarkState],
      flip_signs_for_min: bool = False,
  ) -> convergence_curve.ConvergenceCurve:
    """Generates a ConvergenceCurve from a batch of BenchmarkStates.

    Each state in batch should represent the same study (different repeat).

    Args:
      states: List of BenchmarkStates.
      flip_signs_for_min: If true, flip signs of curve when it is MINIMIZE
        metric.

    Returns:
      Convergence curve with batch size equal to length of states.

    Raises:
      ValueError: When problem statements are not the same or is multiobjective.
    """
    if not states:
      raise ValueError('Empty States.')

    problem_statement = states[0].experimenter.problem_statement()

    curves = []
    for state in states:
      if problem_statement != state.experimenter.problem_statement():
        raise ValueError(
            f'States must have same problem {problem_statement}'
            f' and {state.experimenter.problem_statement()}'
        )
      state_trials = state.algorithm.supporter.GetTrials()

      converter = (
          convergence_curve.MultiMetricCurveConverter.from_metrics_config(
              problem_statement.metric_information,
              flip_signs_for_min=flip_signs_for_min,
          )
      )
      curve = converter.convert(state_trials)
      curves.append(curve)
    return convergence_curve.ConvergenceCurve.align_xs(curves)[0]

  @classmethod
  def to_record(
      cls,
      algorithm: str,
      experimenter_factory: experimenters.SerializableExperimenterFactory,
      states: list[benchmarks.BenchmarkState],
      flip_signs_for_min: bool = False,
  ) -> BenchmarkRecord:
    """Generates a BenchmarkRecord from a batch of BenchmarkStates.

    Each state in batch should represent the same study (different repeat).

    Args:
      algorithm: Algorithm name.
      experimenter_factory: Factory used for running BenchmarkState.
      states: List of BenchmarkStates.
      flip_signs_for_min: If true, flip signs of curve when it is MINIMIZE
        metric.

    Returns:
      BenchmarkRecord.
    """
    plot_elements = {}
    plot_elements[RECORD_OBJECTIVE_KEY] = PlotElement(
        curve=cls.to_curve(states, flip_signs_for_min=flip_signs_for_min),
        yscale='symlog',
    )
    return BenchmarkRecord(
        algorithm=algorithm,
        experimenter_metadata=experimenter_factory.dump(),
        plot_elements=plot_elements,
    )


class BenchmarkRecordAnalyzer:
  """Analyzer for a sequence of Benchmark Records."""

  @classmethod
  def add_comparison_metrics(
      cls,
      records: Sequence[BenchmarkRecord],
      baseline_algo: str,
      *,
      compare_metric: str = RECORD_OBJECTIVE_KEY,
      comparator_factory: convergence_curve.ConvergenceComparatorFactory = convergence_curve.LogEfficiencyConvergenceCurveComparator,
  ) -> list[BenchmarkRecord]:
    """Adds comparison scores as metrics via PlotElements to BenchmarkRecord.

    Comparisons are done for compare_metric with respect to the baseline_algo.

    Args:
      records: Sequence of BenchmarkRecords
      baseline_algo: Baseline algorithm to be compared against.
      compare_metric: Metric of comparison.
      comparator_factory: Comparator used for scoring.

    Returns:
      List of BenchmarkRecords with comparison scores added as metrics.

    Raises:
      ValueError: When baseline_algo cannot be found in records or the metric
      of comparison does not correspond to a curve.
    """
    records_list = [
        (rec.algorithm, json.dumps(dict(rec.experimenter_metadata)), rec)
        for rec in records
    ]
    df = pd.DataFrame(
        records_list, columns=['algorithm', 'experimenter', 'record']
    )

    analyzed_records = []
    for experimenter_key, experimenter_group in df.groupby('experimenter'):
      # Checks and stores the mapping from algorithm to plot_elements
      algo_to_elements_dict = {}
      for algorithm_name, group in experimenter_group.groupby('algorithm'):
        if not group.size:
          continue
        if len(group) != 1:
          output_str = (
              f'Condense {len(group)} records for {algorithm_name} with exptr '
              f' {experimenter_key} before applying comparisons for'
              f' {group}'
          )
          logging.error('%s', output_str)
          continue
        algo_to_elements_dict[algorithm_name] = group.record.iloc[
            0
        ].plot_elements

      # Finds the baseline algorithm and the comparison element.
      if compare_metric not in algo_to_elements_dict[baseline_algo]:
        raise ValueError(
            f'Compare metric {compare_metric} not in baseline {baseline_algo}'
        )
      baseline_element = algo_to_elements_dict[baseline_algo][compare_metric]
      if baseline_element.plot_type != 'error-bar':
        raise ValueError(
            f'No comparison can be done for {compare_metric} since'
            f' plot type is {baseline_element.plot_type}'
        )

      # Attempts to apply comparison and add comparison metrics.
      for algorithm_name, elems_dict in algo_to_elements_dict.items():
        compared_element = elems_dict[compare_metric]
        comparator = comparator_factory(
            baseline_curve=baseline_element.curve,
            compared_curve=compared_element.curve,
        )
        try:
          dict_key = (
              compare_metric + f':{comparator.name}_curve:' + baseline_algo
          )
          elems_dict[dict_key] = PlotElement(
              curve=comparator.curve(),
              plot_type='error-bar',
          )
        except Exception as e:  # pylint: disable=broad-exception-caught
          output_str = (
              f'Skip comparing curve for algo {algorithm_name} with'
              f' {compared_element} \n due to error {e}'
          )
          logging.error('%s', output_str)
        try:
          dict_key = compare_metric + f':{comparator.name}:' + baseline_algo
          elems_dict[dict_key] = PlotElement(
              plot_array=np.asarray([comparator.score()]),
              plot_type='histogram',
          )
        except Exception as e:  # pylint: disable=broad-exception-caught
          output_str = (
              f'Skip comparing score for algo {algorithm_name} with '
              f'  {compared_element} \n due to error {e}'
          )
          logging.error('%s', output_str)
        analyzed_records.append(
            BenchmarkRecord(
                algorithm=algorithm_name,
                experimenter_metadata=vz.Metadata(json.loads(experimenter_key)),  # pytype: disable=wrong-arg-types  # pandas-drop-duplicates-overloads
                plot_elements=elems_dict,
            )
        )

    return analyzed_records



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Algorithms for transfer learning."""

import chex
import equinox as eqx
from jax import numpy as jnp
from tensorflow_probability.substrates import jax as tfp


tfd = tfp.distributions


class TransferPredictionState(eqx.Module):
  """State and metadata for combining this prediction in transfer learning.

  Attributes:
    pred: The prediction to combine
    aux: Auxiliary information about the prediction
    training_data_count: Number of samples to train the `Predictive` that made
      the prediction.
    num_hyperparameters: Hyperparameters of the `Predictive` that made the
      prediction.
  """

  pred: tfd.Distribution
  aux: chex.ArrayTree
  training_data_count: int
  num_hyperparameters: int


def _compute_dof(training_data_count: int, num_hyperparameters: int) -> float:
  """Computes the DOF of a `Predictive`.

  This is a maximum of two measures of DOF.

  The first represents the DOF associated with a log likelihood computation,
  after optimizing the hyperparameters of the kernel, i.e. the
  degrees-of-freedom (dof) of a finite linear regression problem.

  The second represents the fact we know something about the standard
  deviation even when there are more hyperpameters than training data.

  Args:
    training_data_count: Number of samples used to train the `Predictive`.
    num_hyperparameters: Number of hyperparameters in the `Predictive`.

  Returns:
    Returns DOF of the predictive
  """
  return max(
      training_data_count - num_hyperparameters,
      training_data_count / (1 + num_hyperparameters),
  )


def combine_predictions_with_aux(
    top_pred: TransferPredictionState,
    base_pred: TransferPredictionState,
    *,
    expected_base_stddev_mismatch: float = 1.0
) -> tuple[tfd.Distribution, chex.ArrayTree]:
  """Combines two predictions from transfer learning.

  The means are combined as a simple sum.

  The standard deviations are combined using a geometric mean, with a
  weighting coefficient `alpha` that sets their relative importance.

  See the below code for the exact computation of `alpha`, which is a function
  of the uncertainty of the base to the uncertainty of the top-level
  prediction.
  Args:
    top_pred: Prediction from the top model (trained on the base's residuals)
    base_pred: Prediction from the base model (trained on the full data)
    expected_base_stddev_mismatch: Used for combining a base standard deviation
      with the top-level model's standard deviation estimate. Formally, it is
      the expected RMS fractional mismatch in standard deviation between a
      typical base and a typical top-level model (averaged over the feasible
      region). Allowable values are [0, 1] with (0.1, 0.8) being more likely.
      Assumes that the value is allowable, due to compatibility with `jax` and
      avoiding `jax.checkify`. Unexpected results may occur if value is set
      out-of-bounds.

  Returns:
    The combined distribution, assumed to be Normal, and auxiliary information.
  """

  dof_base = _compute_dof(
      training_data_count=base_pred.training_data_count,
      num_hyperparameters=base_pred.num_hyperparameters,
  )
  dof_top = _compute_dof(
      training_data_count=top_pred.training_data_count,
      num_hyperparameters=top_pred.num_hyperparameters,
  )

  # `beta_squared` is the ratio of uncertainty of the base to the uncertainty
  # in the top-level model.  More precisely, it is the
  # variance{ log { stddev returned by the base}} /
  # variance{ log { stddev returned by the top model}}.
  # This is a large number when the top-level stddev is more trustworthy, and
  # small when the base stddev is relatively trustworthy.
  beta_squared = (dof_top / dof_base) * (
      1 + dof_base + (expected_base_stddev_mismatch**2)
  )

  # Finally, compute the geometric mean weight, `alpha`.
  alpha = beta_squared / (1 + beta_squared)

  # Combine the means.
  comb_mean = top_pred.pred.mean() + base_pred.pred.mean()

  # Use `alpha` to combine the stddevs in a weighted geometric mean.
  comb_stddev = jnp.power(top_pred.pred.stddev(), alpha) * jnp.power(
      base_pred.pred.stddev(), (1 - alpha)
  )

  prev_aux = {
      'base_aux': base_pred.aux,
      'top_aux': top_pred.aux,
  }

  # Entries in `aux` must have the same batch shape as the predictions.
  batch_shape = comb_mean.shape[0]
  aux = {
      'prev_aux': prev_aux,
      'mean': comb_mean,
      'stddev': comb_stddev,
      'alpha': jnp.ones(batch_shape) * alpha,
      'expected_base_stddev_mismatch': (
          jnp.ones(batch_shape) * expected_base_stddev_mismatch
      ),
      'beta_squared': jnp.ones(batch_shape) * beta_squared,
  }

  # Assume a multivariate normal distribution with diagonal covariance as output
  return tfd.MultivariateNormalDiag(loc=comb_mean, scale_diag=comb_stddev), aux



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Experimenter factories."""

import abc
import functools
import json
from typing import Optional

import attr
import numpy as np
from vizier import pyvizier as vz
from vizier._src.benchmarks.experimenters import discretizing_experimenter
from vizier._src.benchmarks.experimenters import experimenter
from vizier._src.benchmarks.experimenters import noisy_experimenter
from vizier._src.benchmarks.experimenters import normalizing_experimenter
from vizier._src.benchmarks.experimenters import numpy_experimenter
from vizier._src.benchmarks.experimenters import permuting_experimenter
from vizier._src.benchmarks.experimenters import shifting_experimenter
from vizier._src.benchmarks.experimenters.synthetic import bbob
from vizier.interfaces import serializable
from vizier.utils import json_utils

BBOB_FACTORY_KEY = 'bbob_factory'
SINGLE_OBJECTIVE_FACTORY_KEY = 'single_objective_factory'


class ExperimenterFactory(abc.ABC):
  """Abstraction for creating Experimenters."""

  @abc.abstractmethod
  def __call__(self) -> experimenter.Experimenter:
    """Creates the Experimenter.

    This method should behave deterministcally. That is, multiple invocations of
    __call__ should always return the identical experimenter that given the
    same sequence of trials, behaves exactly the same.

    This method should not be used for generating a random experimenter.

    Returns:
      Same exact experimenter created fresh.
    """


class SerializableExperimenterFactory(
    ExperimenterFactory, serializable.Serializable
):
  """Abstraction for experimenter factories with dump/recover methods.

  Subclasses of this interface can be serialized so that we can re-create
  the exact same experimenter object.
  """


@attr.define
class BBOBExperimenterFactory(SerializableExperimenterFactory):
  """Factory for a BBOB function."""

  # Should be a BBOB function name in bbob.py (name should match exactly).
  name: str = attr.field(default='', validator=attr.validators.instance_of(str))
  dim: int = attr.field(
      default=1,
      validator=[attr.validators.instance_of(int), attr.validators.gt(0)],
  )
  rotation_seed: int = attr.field(default=0)

  def __call__(self) -> numpy_experimenter.NumpyExperimenter:
    bbob_function = getattr(bbob, self.name, None)
    if bbob_function is None:
      raise ValueError(f'{self.name} is not a valid BBOB function in bbob.py')
    bbob_function = functools.partial(bbob_function, seed=self.rotation_seed)
    return numpy_experimenter.NumpyExperimenter(
        bbob_function, bbob.DefaultBBOBProblemStatement(self.dim)
    )

  def dump(self) -> vz.Metadata:
    metadata = vz.Metadata()
    metadata_dict = {
        'name': self.name,
        'dim': self.dim,
        'rotation_seed': self.rotation_seed,
    }
    metadata[BBOB_FACTORY_KEY] = json.dumps(metadata_dict)
    return metadata

  @classmethod
  def recover(cls, metadata: vz.Metadata) -> 'BBOBExperimenterFactory':
    metadata_dict = json.loads(metadata[BBOB_FACTORY_KEY])
    return cls(**metadata_dict)


@attr.define
class SingleObjectiveExperimenterFactory(SerializableExperimenterFactory):
  """Factory for a single objective Experimenter.

  Attributes:
    base_factory:
    shift: An array of doubles that is broadcastable to dim of search space.
    should_restrict: Whether to restrict the parameter bounds of search space
      when shifting. Note that this will change the search space bounds.
    noise_type: Should be one of the noise types in noisy_experimenter.py
    noise_seed: Seed for the noise.
    num_normalization_samples: Number of normalization samples. If zero, no
      normalization is done.
    discrete_dict: Dictionary of parameter indices to discretize in a grid. Key
      = index of parameter to be discretize Value = Number of feasible points to
      discretize to. For example, {0: 3, 2 : 2} discretizes the first parameter
      to 3 feasible points and the third to 2 feasible points. Generally, this
      should be used only when base_factory generates only continuous
      parameters.
    categorical_dict: Dictionary of parameter indices to categorize in a grid.
      Similar to `discrete_dict`, except this converts a parameter to
      categorical instead of discrete.
    permute_categoricals: Whether to permute the categorical values.
  """

  base_factory: SerializableExperimenterFactory = attr.field()
  shift: Optional[np.ndarray] = attr.field(default=None)
  should_restrict: bool = attr.field(default=True, kw_only=True)
  noise_type: Optional[str] = attr.field(default=None)
  noise_seed: int = attr.field(default=0)
  num_normalization_samples: int = attr.field(default=0)
  discrete_dict: dict[int, int] = attr.field(default=attr.Factory(dict))
  categorical_dict: dict[int, int] = attr.field(default=attr.Factory(dict))
  permute_categoricals: bool = attr.field(default=True, kw_only=True)
  # TODO: Add support for sparsification.

  def __call__(self) -> experimenter.Experimenter:
    """Creates the SingleObjective Experimenter."""
    exptr = self.base_factory()
    if self.shift is not None:
      exptr = shifting_experimenter.ShiftingExperimenter(
          exptr, shift=self.shift, should_restrict=self.should_restrict
      )
    if self.num_normalization_samples:
      exptr = normalizing_experimenter.NormalizingExperimenter(
          exptr, num_normalization_samples=self.num_normalization_samples
      )

    # Discretization and categorization.
    if self.discrete_dict.keys() & self.categorical_dict.keys():
      raise ValueError(
          f'{self.discrete_dict} discretizing indices overlap with '
          f'{self.categorical_dict} categorical indices'
      )

    pcs = list(exptr.problem_statement().search_space.parameters)
    if self.discrete_dict:
      discretization = {
          pcs[idx].name: points for idx, points in self.discrete_dict.items()
      }
      exptr = (
          discretizing_experimenter.DiscretizingExperimenter.create_with_grid(
              exptr, discretization, convert_to_str=False
          )
      )

    if self.categorical_dict:
      categorization = {
          pcs[idx].name: points for idx, points in self.categorical_dict.items()
      }
      exptr = (
          discretizing_experimenter.DiscretizingExperimenter.create_with_grid(
              exptr, categorization, convert_to_str=True
          )
      )

    if self.permute_categoricals:
      search_space = exptr.problem_statement().search_space
      categorical_parameters = [
          p.name
          for p in search_space.parameters
          if p.type == vz.ParameterType.CATEGORICAL
      ]
      exptr = permuting_experimenter.PermutingExperimenter(
          exptr, categorical_parameters
      )

    if self.noise_type is not None:
      exptr = noisy_experimenter.NoisyExperimenter.from_type(
          exptr, noise_type=self.noise_type.upper(), seed=self.noise_seed
      )

    return exptr

  def dump(self) -> vz.Metadata:
    # The resulting metadata stores base factory metadata
    # and metadata_dict with different keys.
    metadata = self.base_factory.dump()
    metadata_dict = {
        'shift': self.shift,
        'noise_type': self.noise_type,
        'num_normalization_samples': self.num_normalization_samples,
        'discrete_dict': self.discrete_dict,
        'categorical_dict': self.categorical_dict,
        'should_restrict': self.should_restrict,
        'permute_categoricals': self.permute_categoricals,
        'noise_seed': self.noise_seed,
    }
    metadata[SINGLE_OBJECTIVE_FACTORY_KEY] = json.dumps(
        metadata_dict, cls=json_utils.NumpyEncoder
    )
    return metadata

  @classmethod
  def recover(
      cls, metadata: vz.Metadata
  ) -> 'SingleObjectiveExperimenterFactory':
    if BBOB_FACTORY_KEY in metadata:
      base_factory = BBOBExperimenterFactory.recover(metadata)
    else:
      raise serializable.DecodeError(
          f'No valid base factory found in {metadata}'
      )

    metadata_dict = json.loads(
        metadata[SINGLE_OBJECTIVE_FACTORY_KEY], cls=json_utils.NumpyDecoder
    )

    # Turn string keys back to int for discrete/categorical dicts.
    int_discrete_dict = {
        int(k): v for k, v in metadata_dict['discrete_dict'].items()
    }
    metadata_dict['discrete_dict'] = int_discrete_dict
    int_categorical_dict = {
        int(k): v for k, v in metadata_dict['categorical_dict'].items()
    }
    metadata_dict['categorical_dict'] = int_categorical_dict

    return SingleObjectiveExperimenterFactory(
        base_factory=base_factory, **metadata_dict
    )
